"""Filesystem tools for file operations."""

from typing import Optional
from open_agent.agents.tools.base import BaseTool, PermissionLevel
from open_agent.agents.sandbox.backend import OpenShellBackend, SandboxSession
import structlog

logger = structlog.get_logger(__name__)


class FileTools(BaseTool):
    """File system tools (ls, read, write, edit).
    
    Provides safe filesystem operations within a sandboxed environment,
    similar to OpenClaw's filesystem tools.
    
    Example:
        >>> backend = OpenShellBackend()
        >>> tools = FileTools(backend=backend)
        >>> content = await tools.read_file("/path/to/file")
        >>> await tools.write_file("/path/to/file", "content")
    """

    def __init__(self, backend: Optional[OpenShellBackend] = None):
        super().__init__(
            name="filesystem",
            description="File system operations",
            permission_level=PermissionLevel.WRITE,
        )
        self.backend = backend
        self.session: Optional[SandboxSession] = None

    def _get_session(self) -> SandboxSession:
        """Get or create a sandbox session."""
        if self.session is None:
            if self.backend is None:
                self.backend = OpenShellBackend()
            self.session = self.backend.create_sandbox("nemotron")
        return self.session

    async def ls(self, path: str = ".") -> str:
        """List directory contents.
        
        Args:
            path: Directory path to list
            
        Returns:
            Directory listing
        """
        session = self._get_session()
        result = session.exec(f"ls -la {path}")
        return result.get("stdout", "")

    async def read_file(self, path: str) -> str:
        """Read a file.
        
        Args:
            path: File path to read
            
        Returns:
            File contents or error
        """
        session = self._get_session()
        return session.read_file(path)

    async def write_file(self, path: str, content: str) -> str:
        """Write a file.
        
        Args:
            path: File path to write
            content: Content to write
            
        Returns:
            Success or error message
        """
        session = self._get_session()
        result = session.write_file(path, content)
        if result.get("success"):
            return f"File written successfully: {path}"
        else:
            return f"Error: {result.get('error', 'Unknown error')}"

    async def edit_file(self, path: str, old_text: str, new_text: str) -> str:
        """Edit a file (replace old_text with new_text).
        
        Args:
            path: File path to edit
            old_text: Text to replace
            new_text: Replacement text
            
        Returns:
            Success or error message
        """
        content = await self.read_file(path)
        content = content.replace(old_text, new_text)
        return await self.write_file(path, content)



"""Channel registry - Manages available channels."""

from typing import Dict, List, Optional, Any, Callable
from dataclasses import dataclass
import structlog

logger = structlog.get_logger(__name__)


@dataclass
class Channel:
    """A communication channel."""
    name: str
    description: str
    enabled: bool = True
    config: Optional[Dict[str, Any]] = None
    handler: Optional[Callable] = None


class ChannelRegistry:
    """Registry for managing communication channels.
    
    Similar to OpenClaw's channel registry for managing
    integrations with various messaging platforms.
    """

    def __init__(self):
        self.channels: Dict[str, Channel] = {}

    def register(
        self,
        name: str,
        description: str,
        handler: Optional[Callable] = None,
        config: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Register a channel."""
        self.channels[name] = Channel(
            name=name,
            description=description,
            handler=handler,
            config=config,
        )
        logger.info("channel_registered", name=name)

    def unregister(self, name: str) -> bool:
        """Unregister a channel."""
        if name in self.channels:
            del self.channels[name]
            logger.info("channel_unregistered", name=name)
            return True
        return False

    def get(self, name: str) -> Optional[Channel]:
        """Get a channel by name."""
        return self.channels.get(name)

    def list(self, enabled_only: bool = True) -> List[Channel]:
        """List all channels."""
        channels = list(self.channels.values())
        if enabled_only:
            channels = [c for c in channels if c.enabled]
        return channels

    def enable(self, name: str) -> bool:
        """Enable a channel."""
        channel = self.channels.get(name)
        if channel:
            channel.enabled = True
            return True
        return False

    def disable(self, name: str) -> bool:
        """Disable a channel."""
        channel = self.channels.get(name)
        if channel:
            channel.enabled = False
            return True
        return False



"""Configuration management for OpenAgent."""

from typing import Optional
from pydantic import Field
from functools import lru_cache


class NVIDIAConfig:
    """NVIDIA API configuration."""
    def __init__(
        self,
        api_key: str = "",
        base_url: str = "https://integrate.api.nvidia.com/v1",
        model: str = "nvidia/nemotron-3-super-120b-a12b",
    ):
        import os
        self.api_key = api_key or os.environ.get("NVIDIA_API_KEY", "")
        self.base_url = base_url
        self.model = model


class OpenShellConfig:
    """OpenShell sandbox configuration."""
    def __init__(
        self,
        enabled: bool = True,
        sandbox_type: str = "auto",
        policy_file: Optional[str] = None,
        credential_provider: str = "nvidia",
        remote_host: Optional[str] = None,
    ):
        import os
        self.enabled = enabled
        self.sandbox_type = sandbox_type
        self.policy_file = policy_file
        self.credential_provider = credential_provider
        self.remote_host = remote_host or os.environ.get("OPENSHELL_REMOTE_HOST")


class AgentConfig:
    """Agent configuration."""
    def __init__(
        self,
        name: str = "open-agent",
        model: str = "nvidia/nemotron-3-super-120b-a12b",
        model_provider: str = "nvidia",
        max_iterations: int = 50,
        max_retries: int = 6,
        timeout: int = 300,
        temperature: float = 0.7,
        system_prompt: Optional[str] = None,
    ):
        import os
        self.name = name
        self.model = model or os.environ.get("AGENT_MODEL", model)
        self.model_provider = model_provider
        self.max_iterations = max_iterations
        self.max_retries = max_retries
        self.timeout = timeout
        self.temperature = temperature
        self.system_prompt = system_prompt


class MemoryConfig:
    """Memory persistence configuration."""
    def __init__(
        self,
        enabled: bool = True,
        backend: str = "store",
        storage_path: str = "./memory",
        max_long_term: int = 1000,
    ):
        self.enabled = enabled
        self.backend = backend
        self.storage_path = storage_path
        self.max_long_term = max_long_term


class OpenAgentConfig:
    """Main OpenAgent configuration."""
    def __init__(
        self,
        nvidia: Optional[NVIDIAConfig] = None,
        openshell: Optional[OpenShellConfig] = None,
        agent_config: Optional[AgentConfig] = None,
        memory: Optional[MemoryConfig] = None,
    ):
        self.nvidia = nvidia or NVIDIAConfig()
        self.openshell = openshell or OpenShellConfig()
        self.agent_config = agent_config or AgentConfig()
        self.memory = memory or MemoryConfig()


_config_instance: Optional[OpenAgentConfig] = None


@lru_cache
def get_config() -> OpenAgentConfig:
    """Get cached configuration instance."""
    global _config_instance
    if _config_instance is None:
        _config_instance = OpenAgentConfig()
    return _config_instance


def load_config_from_file(config_path: str) -> OpenAgentConfig:
    """Load configuration from a YAML file."""
    import yaml
    with open(config_path, "r") as f:
        data = yaml.safe_load(f)
    return OpenAgentConfig(**data)



"""Tool registry and tool base class."""

from typing import Any, Callable
from pydantic import BaseModel, Field
import structlog

from open_agent.core.types import ToolCall, ToolResult, PermissionLevel

logger = structlog.get_logger(__name__)


class Tool(BaseModel):
    """Base tool definition."""
    name: str
    description: str
    parameters: dict[str, Any] = Field(default_factory=dict)
    permission_level: PermissionLevel = PermissionLevel.EXEC
    is_async: bool = True

    def execute(self, **kwargs) -> Any:
        """Execute the tool synchronously."""
        raise NotImplementedError

    async def aexecute(self, **kwargs) -> Any:
        """Execute the tool asynchronously."""
        return self.execute(**kwargs)


class ToolMetadata(BaseModel):
    """Metadata about a registered tool."""
    name: str
    description: str
    permission_level: PermissionLevel
    is_async: bool


class ToolRegistry:
    """Registry for managing and accessing tools."""

    def __init__(self):
        self._tools: dict[str, Tool] = {}
        self._permissions: dict[str, PermissionLevel] = {}

    def register(self, tool: Tool) -> None:
        """Register a tool."""
        self._tools[tool.name] = tool
        self._permissions[tool.name] = tool.permission_level
        logger.info("tool_registered", tool_name=tool.name)

    def unregister(self, name: str) -> bool:
        """Unregister a tool."""
        if name in self._tools:
            del self._tools[name]
            del self._permissions[name]
            logger.info("tool_unregistered", tool_name=name)
            return True
        return False

    def get(self, name: str) -> Tool | None:
        """Get a tool by name."""
        return self._tools.get(name)

    def list(self) -> list[ToolMetadata]:
        """List all registered tools."""
        return [
            ToolMetadata(
                name=name,
                description=tool.description,
                permission_level=tool.permission_level,
                is_async=tool.is_async,
            )
            for name, tool in self._tools.items()
        ]

    def check_permission(self, name: str, required_level: PermissionLevel) -> bool:
        """Check if a tool has the required permission level."""
        tool_level = self._permissions.get(name)
        if not tool_level:
            return False
        levels = list(PermissionLevel)
        return levels.index(tool_level) >= levels.index(required_level)

    async def execute(self, name: str, arguments: dict[str, Any]) -> ToolResult:
        """Execute a tool by name."""
        tool = self._tools.get(name)
        if not tool:
            return ToolResult(
                tool_call_id="",
                success=False,
                error=f"Tool '{name}' not found",
            )

        tool_call_id = f"call_{id(tool)}"
        try:
            if tool.is_async:
                result = await tool.aexecute(**arguments)
            else:
                result = tool.execute(**arguments)
            return ToolResult(
                tool_call_id=tool_call_id,
                success=True,
                output=str(result),
            )
        except Exception as e:
            logger.error("tool_execution_failed", tool_name=name, error=str(e))
            return ToolResult(
                tool_call_id=tool_call_id,
                success=False,
                error=str(e),
            )

    def has_tool(self, name: str) -> bool:
        """Check if a tool is registered."""
        return name in self._tools


# Global default registry
default_registry = ToolRegistry()



"""Authentication handler for gateway requests."""

from typing import Optional, Dict, Any, List
from dataclasses import dataclass
from datetime import datetime, timedelta
import structlog

logger = structlog.get_logger(__name__)


@dataclass
class AuthCredentials:
    """Authentication credentials."""
    user_id: str
    token: Optional[str] = None
    api_key: Optional[str] = None
    expires_at: Optional[datetime] = None


class AuthHandler:
    """Handles authentication for gateway requests.
    
    Similar to OpenClaw's gateway auth system for securing
    access to the agent platform.
    """

    def __init__(self, secret_key: Optional[str] = None):
        self.secret_key = secret_key or "default-secret-key"
        self.credentials: Dict[str, AuthCredentials] = {}
        self.rate_limits: Dict[str, List[datetime]] = {}

    def authenticate(
        self,
        user_id: str,
        token: Optional[str] = None,
        api_key: Optional[str] = None,
    ) -> bool:
        """Authenticate a user."""
        creds = self.credentials.get(user_id)
        if not creds:
            return False

        if creds.expires_at and datetime.now() > creds.expires_at:
            return False

        if token and creds.token != token:
            return False

        if api_key and creds.api_key != api_key:
            return False

        logger.info("user_authenticated", user_id=user_id)
        return True

    def authorize(
        self,
        user_id: str,
        required_permissions: List[str],
    ) -> bool:
        """Check if user has required permissions."""
        return user_id in self.credentials

    def add_credentials(
        self,
        user_id: str,
        token: Optional[str] = None,
        api_key: Optional[str] = None,
        expires_in: Optional[int] = None,
    ) -> None:
        """Add credentials for a user."""
        expires_at = None
        if expires_in:
            expires_at = datetime.now() + timedelta(seconds=expires_in)

        self.credentials[user_id] = AuthCredentials(
            user_id=user_id,
            token=token,
            api_key=api_key,
            expires_at=expires_at,
        )

    def remove_credentials(self, user_id: str) -> bool:
        """Remove credentials for a user."""
        if user_id in self.credentials:
            del self.credentials[user_id]
            return True
        return False

    def check_rate_limit(
        self,
        user_id: str,
        max_requests: int = 100,
        window_seconds: int = 60,
    ) -> bool:
        """Check if user is within rate limits."""
        now = datetime.now()
        cutoff = now - timedelta(seconds=window_seconds)

        if user_id not in self.rate_limits:
            self.rate_limits[user_id] = []

        self.rate_limits[user_id] = [
            ts for ts in self.rate_limits[user_id] if ts > cutoff
        ]

        if len(self.rate_limits[user_id]) >= max_requests:
            return False

        self.rate_limits[user_id].append(now)
        return True



"""Todo tool for planning and task tracking."""

from typing import List, Dict, Any
from open_agent.agents.tools.base import BaseTool, PermissionLevel
import structlog

logger = structlog.get_logger(__name__)


class TodoTool(BaseTool):
    """Todo list tool for planning and task tracking.
    
    Provides task management capabilities for complex multi-step tasks,
    similar to OpenClaw's todo tool.
    
    Example:
        >>> tool = TodoTool()
        >>> await tool.write_todos([
        ...     {"id": "1", "content": "Task 1", "status": "pending"},
        ...     {"id": "2", "content": "Task 2", "status": "pending"},
        ... ])
        >>> todos = await tool.get_todos()
    """

    def __init__(self):
        super().__init__(
            name="todo",
            description="Track and manage todo items for complex tasks",
            permission_level=PermissionLevel.WRITE,
        )
        self.todos: List[Dict[str, Any]] = []

    async def execute(self, action: str, **kwargs) -> str:
        """Execute a todo action.
        
        Args:
            action: Action to perform (get, write, complete)
            **kwargs: Action-specific arguments
            
        Returns:
            Action result
        """
        if action == "get":
            return str(await self.get_todos())
        elif action == "write":
            todos = kwargs.get("todos", [])
            return await self.write_todos(todos)
        elif action == "complete":
            todo_id = kwargs.get("todo_id", "")
            return await self.complete_todo(todo_id)
        else:
            return f"Unknown action: {action}"

    async def write_todos(self, todos: List[Dict[str, Any]]) -> str:
        """Write/update todo list.
        
        Args:
            todos: List of todo items
            
        Returns:
            Status message
        """
        self.todos = todos
        return f"Updated {len(todos)} todo items"

    async def complete_todo(self, todo_id: str) -> str:
        """Mark a todo as complete.
        
        Args:
            todo_id: ID of the todo to complete
            
        Returns:
            Status message
        """
        for todo in self.todos:
            if todo.get("id") == todo_id:
                todo["status"] = "complete"
                return f"Marked todo '{todo.get('content', todo_id)}' as complete"
        return f"Todo '{todo_id}' not found"

    async def get_todos(self) -> List[Dict[str, Any]]:
        """Get current todo list.
        
        Returns:
            List of todo items
        """
        return self.todos



#!/usr/bin/env python3
"""
🚀 SmithAI Production Crews - Real-World Multi-Agent Workflows

This example demonstrates production-ready agent crews that solve real problems:
1. Code Review Crew - PR review with automated feedback
2. DevOps Crew - CI/CD pipeline automation
3. Content Crew - Research → Write → Edit pipeline

Run with: python examples/05_production_crews.py
"""

import asyncio
import sys
from open_agent import create_crew


# =============================================================================
# CREW 1: Code Review Automation
# =============================================================================

async def run_code_review_crew(code_snippet: str, language: str = "python"):
    """Automated code review with expert feedback."""
    
    print("\n" + "="*60)
    print("🚀 CREW 1: Code Review Automation")
    print("="*60)
    
    crew = create_crew(
        agents_config=[
            {
                "name": "analyzer",
                "role": "Code Analyzer",
                "goal": "Analyze code for bugs, performance issues, and best practices",
                "backstory": """You are a senior code reviewer with 20 years of experience.
                You excel at finding bugs, performance bottlenecks, security vulnerabilities,
                and code quality issues. You know SOLID principles, design patterns,
                and industry best practices for multiple languages."""
            },
            {
                "name": "reviewer",
                "role": "Code Reviewer",
                "goal": "Provide constructive, actionable feedback on code",
                "backstory": """You are a tech lead who provides clear, actionable feedback.
                You balance thoroughness with practicality, focusing on issues that
                matter most for maintainability and performance."""
            },
            {
                "name": "fixer",
                "role": "Code Fixer",
                "goal": "Write improved versions of code with fixes applied",
                "backstory": """You are a senior developer who writes clean, efficient code.
                You apply fixes while maintaining the original intent and functionality."""
            },
        ],
        tasks_config=[
            {
                "description": f"Analyze this {language} code and identify all issues: bugs, performance problems, security vulnerabilities, SOLID violations, and best practice violations. Format your response as a numbered list.\n\nCode:\n```{language}\n{code_snippet}\n```",
                "agent": "analyzer",
                "expected_output": "Detailed list of all code issues with severity levels"
            },
            {
                "description": "Review the issues found and prioritize them. Focus on the top 5 most critical issues that MUST be fixed. Provide specific, actionable recommendations for each.",
                "agent": "reviewer",
                "expected_output": "Prioritized list of top 5 critical issues with recommendations"
            },
            {
                "description": "Based on the feedback, write an improved version of the code that fixes the critical issues. Maintain the original functionality while improving quality.",
                "agent": "fixer",
                "expected_output": "Improved code with fixes explained in comments"
            },
        ],
        process="sequential",
        verbose=True,
    )
    
    results = await crew.kickoff()
    return results


# =============================================================================
# CREW 2: DevOps Automation
# =============================================================================

async def run_devops_crew(task: str):
    """CI/CD and DevOps task automation."""
    
    print("\n" + "="*60)
    print("🚀 CREW 2: DevOps Automation")
    print("="*60)
    
    crew = create_crew(
        agents_config=[
            {
                "name": "architect",
                "role": "Infrastructure Architect",
                "goal": "Design scalable, reliable infrastructure solutions",
                "backstory": """You are a DevOps architect with expertise in Kubernetes, Docker,
                AWS/GCP/Azure, Terraform, and CI/CD pipelines. You design systems that
                are scalable, reliable, and cost-effective."""
            },
            {
                "name": "implementer",
                "role": "DevOps Engineer",
                "goal": "Implement infrastructure as code and automation scripts",
                "backstory": """You are a DevOps engineer who writes production-ready
                Terraform, Docker, Kubernetes, and CI/CD configurations.
                Your code is always production-quality with proper error handling."""
            },
            {
                "name": "doc_writer",
                "role": "Technical Writer",
                "goal": "Document infrastructure and processes clearly",
                "backstory": """You write clear, comprehensive documentation that
                helps teams understand and maintain infrastructure."""
            },
        ],
        tasks_config=[
            {
                "description": f"Analyze this DevOps task and design an architecture solution: {task}\n\nProvide: 1) Architecture diagram (text-based), 2) Technology stack recommendations, 3) Estimated setup complexity",
                "agent": "architect",
                "expected_output": "Architecture design with recommendations"
            },
            {
                "description": "Based on the architecture, implement the infrastructure as code. Write Terraform/Docker/Kubernetes configurations or CI/CD scripts as needed.",
                "agent": "implementer",
                "expected_output": "Production-ready infrastructure code"
            },
            {
                "description": "Create a README with setup instructions, architecture overview, and usage guide for the implemented solution.",
                "agent": "doc_writer",
                "expected_output": "Complete documentation"
            },
        ],
        process="sequential",
        verbose=True,
    )
    
    results = await crew.kickoff()
    return results


# =============================================================================
# CREW 3: Content Pipeline
# =============================================================================

async def run_content_crew(topic: str):
    """Automated content creation pipeline."""
    
    print("\n" + "="*60)
    print("🚀 CREW 3: Content Creation Pipeline")
    print("="*60)
    
    crew = create_crew(
        agents_config=[
            {
                "name": "researcher",
                "role": "Content Researcher",
                "goal": "Research topics thoroughly and gather key information",
                "backstory": """You are a research expert who finds the most relevant,
                accurate, and up-to-date information on any topic. You know how to
                identify credible sources and synthesize key points."""
            },
            {
                "name": "writer",
                "role": "Content Writer",
                "goal": "Create engaging, well-structured content",
                "backstory": """You are an expert content writer who creates engaging,
                SEO-friendly content. You write clear headlines, compelling introductions,
                and actionable conclusions."""
            },
            {
                "name": "editor",
                "role": "Senior Editor",
                "goal": "Polish content to publication quality",
                "backstory": """You are a meticulous editor who ensures content is
                error-free, well-formatted, and ready for publication. You check
                grammar, style, and consistency."""
            },
        ],
        tasks_config=[
            {
                "description": f"Research this topic and gather key information: {topic}\n\nProvide: 1) Overview, 2) 5 key points, 3) Real-world examples, 4) Statistics or data if available",
                "agent": "researcher",
                "expected_output": "Comprehensive research summary"
            },
            {
                "description": "Write a well-structured article based on the research. Include: catchy headline, intro, 3-4 main sections with subheadings, and a conclusion with call-to-action.",
                "agent": "writer",
                "expected_output": "Complete article draft"
            },
            {
                "description": "Edit the article for: grammar, clarity, flow, SEO optimization, and consistency. Add meta description and tags.",
                "agent": "editor",
                "expected_output": "Publication-ready article"
            },
        ],
        process="sequential",
        verbose=True,
    )
    
    results = await crew.kickoff()
    return results


# =============================================================================
# MAIN MENU
# =============================================================================

async def main():
    print("""
╔══════════════════════════════════════════════════════════════╗
║                                                              ║
║   🚀 SmithAI Production Crews - Real-World Workflows         ║
║                                                              ║
║   Choose a crew to run:                                      ║
║                                                              ║
║   [1] Code Review - Analyze and fix code issues             ║
║   [2] DevOps - Infrastructure automation                     ║
║   [3] Content - Research → Write → Edit pipeline             ║
║   [4] Run all crews (demo mode)                              ║
║   [Q] Quit                                                   ║
║                                                              ║
╚══════════════════════════════════════════════════════════════╝
    """)
    
    choice = input("Enter choice: ").strip().lower()
    
    if choice == "1":
        code = '''def calculate_total(items):
    total = 0
    for i in items:
        total += i['price'] * i['quantity']
    return total

def process_order(order):
    total = calculate_total(order['items'])
    if total > 1000:
        total = total * 0.9  # 10% discount
    print("Order processed")
    return total'''
        
        results = await run_code_review_crew(code, "python")
        
        print("\n" + "="*60)
        print("📄 FINAL ARTICLE")
        print("="*60)
        for task, result in results.items():
            print(f"\n### {task[:50]}...\n{result}\n")
    
    elif choice == "2":
        results = await run_devops_crew("Set up a production Kubernetes cluster on AWS with auto-scaling, monitoring, and CI/CD")
        
        print("\n" + "="*60)
        print("📄 DEVOPS OUTPUT")
        print("="*60)
        for task, result in results.items():
            print(f"\n### {task[:50]}...\n{result}\n")
    
    elif choice == "3":
        results = await run_content_crew("The future of AI agents in software development")
        
        print("\n" + "="*60)
        print("📄 FINAL ARTICLE")
        print("="*60)
        print(results[list(results.keys())[-1]])
    
    elif choice == "4":
        print("\n🔄 Running demo mode...")
        
        # Quick demo of each crew
        results1 = await run_code_review_crew("x = input(); print(x)", "python")
        print("\n✅ Code review complete!\n")
        
        results2 = await run_content_crew("Why Python is great for AI")
        print("\n✅ Content pipeline complete!\n")
    
    elif choice == "q":
        print("Goodbye!")
        return
    
    else:
        print("Invalid choice")


if __name__ == "__main__":
    asyncio.run(main())



"""Core module for OpenAgent."""

from open_agent.core.gateway import Gateway, create_gateway, Channel, Binding, Session
from open_agent.core.types import (
    BaseAgent,
    AgentConfig,
    AgentContext,
    Message,
    ToolCall,
    ToolResult,
    PermissionLevel,
    SessionState,
)

__all__ = [
    "Gateway",
    "create_gateway",
    "Channel",
    "Binding",
    "Session",
    "BaseAgent",
    "AgentConfig",
    "AgentContext",
    "Message",
    "ToolCall",
    "ToolResult",
    "PermissionLevel",
    "SessionState",
]



"""Base transport classes for channel communication."""

from typing import Optional, Dict, Any, Callable, Awaitable
from dataclasses import dataclass
import asyncio
import structlog

logger = structlog.get_logger(__name__)


@dataclass
class TransportConfig:
    """Configuration for a transport."""
    name: str
    host: str = "0.0.0.0"
    port: int = 8080
    ssl_enabled: bool = False
    ssl_cert: Optional[str] = None
    ssl_key: Optional[str] = None
    max_connections: int = 1000
    timeout: int = 30
    metadata: Dict[str, Any] = None

    def __post_init__(self):
        if self.metadata is None:
            self.metadata = {}


class Transport:
    """Base class for transport implementations.
    
    Similar to OpenClaw's transport layer for handling
    incoming and outgoing communications.
    """

    def __init__(self, config: TransportConfig):
        self.config = config
        self._running = False
        self._handlers: Dict[str, Callable] = {}

    async def start(self) -> None:
        """Start the transport."""
        if self._running:
            return
        self._running = True
        logger.info("transport_started", name=self.config.name)

    async def stop(self) -> None:
        """Stop the transport."""
        if not self._running:
            return
        self._running = False
        logger.info("transport_stopped", name=self.config.name)

    def register_handler(self, event: str, handler: Callable) -> None:
        """Register an event handler."""
        self._handlers[event] = handler

    async def send(self, target: str, message: Any) -> bool:
        """Send a message to a target."""
        raise NotImplementedError

    async def broadcast(self, message: Any) -> None:
        """Broadcast a message to all connected clients."""
        raise NotImplementedError

    @property
    def is_running(self) -> bool:
        """Check if transport is running."""
        return self._running



"""Shell tool for executing commands in sandbox."""

from typing import Optional
from open_agent.agents.tools.base import BaseTool, PermissionLevel
from open_agent.agents.sandbox.backend import OpenShellBackend, SandboxSession
import structlog

logger = structlog.get_logger(__name__)


class ShellTool(BaseTool):
    """Tool for shell command execution via OpenShell sandbox.
    
    Provides secure shell command execution within a sandboxed environment,
    similar to OpenClaw's exec tool.
    
    Example:
        >>> backend = OpenShellBackend()
        >>> tool = ShellTool(backend=backend)
        >>> result = await tool.execute(command="ls -la")
    """

    def __init__(self, backend: Optional[OpenShellBackend] = None):
        super().__init__(
            name="execute",
            description="Execute a shell command in the sandbox",
            permission_level=PermissionLevel.EXEC,
        )
        self.backend = backend
        self.session: Optional[SandboxSession] = None

    async def execute(self, command: str, **kwargs) -> str:
        """Execute a shell command.
        
        Args:
            command: The shell command to execute
            **kwargs: Additional arguments (ignored)
            
        Returns:
            Command output or error message
        """
        if self.session is None:
            if self.backend is None:
                self.backend = OpenShellBackend()
            self.session = self.backend.create_sandbox("nemotron")

        result = self.session.exec(command)
        if result.get("success"):
            return result.get("stdout", "")
        else:
            return f"Error: {result.get('stderr', 'Unknown error')}"



"""Canvas module - Interactive code editing workspace.

This module provides canvas functionality similar to OpenClaw:
- Interactive code editing
- Multi-file workspace
- Syntax highlighting
- Code execution
"""

from open_agent.canvas.workspace import Canvas, CanvasFile, CanvasSession, create_canvas

__all__ = [
    "Canvas",
    "CanvasFile",
    "CanvasSession",
    "create_canvas",
]



"""NVIDIA Nemotron model provider.

This module provides integration with NVIDIA Nemotron models via
langchain-nvidia-ai-endpoints.
"""

from typing import Optional, Any, List
import structlog
import os

logger = structlog.get_logger(__name__)


class NemotronProvider:
    """NVIDIA Nemotron model provider.
    
    Provides access to NVIDIA Nemotron models through the langchain library.
    Supports various Nemotron models including:
    - nemotron-3-super-120b-a12b
    - nemotron-4-340b-a12b
    - nemotron-4-mini-a12b
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        model_name: str = "nvidia/nemotron-3-super-120b-a12b",
        temperature: float = 0.7,
        max_tokens: int = 4096,
    ):
        self.api_key = api_key or os.environ.get("NVIDIA_API_KEY")
        self.model_name = model_name
        self.temperature = temperature
        self.max_tokens = max_tokens
        self.model = None

    def initialize(self) -> None:
        """Initialize the model."""
        if self.model is not None:
            return

        try:
            from langchain_nvidia_ai_endpoints import ChatNVIDIA
            
            self.model = ChatNVIDIA(
                model=self.model_name,
                api_key=self.api_key,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
            )
            logger.info("nemotron_initialized", model=self.model_name)
        except ImportError:
            logger.error("langchain_nvidia_ai_endpoints_not_installed")
            raise ImportError(
                "langchain-nvidia-ai-endpoints is required. "
                "Install with: pip install langchain-nvidia-ai-endpoints"
            )

    async def invoke(self, messages: List[Any]) -> Any:
        """Invoke the model with messages."""
        if self.model is None:
            self.initialize()

        return await self.model.ainvoke(messages)

    async def ainvoke(self, messages: List[Any]) -> Any:
        """Async invoke the model with messages."""
        return await self.invoke(messages)

    def invoke_sync(self, messages: List[Any]) -> Any:
        """Synchronously invoke the model."""
        if self.model is None:
            self.initialize()

        return self.model.invoke(messages)


_global_provider: Optional[NemotronProvider] = None


def create_nemotron_model(
    model_name: str = "nvidia/nemotron-3-super-120b-a12b",
    temperature: float = 0.7,
    api_key: Optional[str] = None,
) -> NemotronProvider:
    """Create or get the global Nemotron provider."""
    global _global_provider
    
    if _global_provider is None:
        _global_provider = NemotronProvider(
            model_name=model_name,
            temperature=temperature,
            api_key=api_key,
        )
        _global_provider.initialize()
    
    return _global_provider


def get_nemotron_model() -> Optional[NemotronProvider]:
    """Get the global Nemotron provider if available."""
    return _global_provider



"""Tools package."""

from open_agent.agents.tools.base import (
    BaseTool,
    ToolResult,
    ToolCall,
    ToolRegistry,
    PermissionLevel,
)
from open_agent.agents.tools.shell import ShellTool
from open_agent.agents.tools.filesystem import FileTools
from open_agent.agents.tools.todo import TodoTool

__all__ = [
    "BaseTool",
    "ToolResult",
    "ToolCall",
    "ToolRegistry",
    "PermissionLevel",
    "ShellTool",
    "FileTools",
    "TodoTool",
]



"""Google Workspace integration - Official Google APIs.

Real API integration with Google Workspace using official google-api-python-client,
google-auth, and google-oauthlib libraries. Matches OpenClaw's Google integration.
"""

import os
import asyncio
from typing import Optional, Dict, List, Any, Callable
from dataclasses import dataclass
from pathlib import Path
import json
import structlog

logger = structlog.get_logger(__name__)


@dataclass
class GoogleOAuthConfig:
    """Google OAuth2 configuration."""
    client_id: str = ""
    client_secret: str = ""
    redirect_uri: str = "http://localhost:8080/oauth/callback"
    scopes: List[str] = None

    def __post_init__(self):
        if self.scopes is None:
            self.scopes = [
                "https://www.googleapis.com/auth/gmail.readonly",
                "https://www.googleapis.com/auth/gmail.send",
                "https://www.googleapis.com/auth/calendar.readonly",
                "https://www.googleapis.com/auth/calendar.events",
                "https://www.googleapis.com/auth/drive.readonly",
                "https://www.googleapis.com/auth/drive.file",
            ]


@dataclass
class GoogleTokens:
    """OAuth tokens storage."""
    token: str = ""
    refresh_token: str = ""
    token_uri: str = "https://oauth2.googleapis.com/token"
    client_id: str = ""
    client_secret: str = ""
    scopes: List[str] = None
    expiry: float = 0


class GoogleOAuthFlow:
    """Google OAuth2 authentication flow - matches OpenClaw implementation."""

    def __init__(self, config: GoogleOAuthConfig):
        self.config = config
        self._flow = None
        self._creds = None

    def get_authorization_url(self) -> tuple[str, str]:
        """Get authorization URL and state token."""
        try:
            from google_auth_oauthlib.flow import Flow

            self._flow = Flow.from_client_config(
                {
                    "web": {
                        "client_id": self.config.client_id,
                        "client_secret": self.config.client_secret,
                        "redirect_uris": [self.config.redirect_uri],
                    }
                },
                scopes=self.config.scopes,
            )
            self._flow.redirect_uri = self.config.redirect_uri

            auth_url, state = self._flow.authorization_url(
                access_type="offline",
                prompt="consent",
                include_granted_scopes="true",
            )

            logger.info("google_auth_url_generated")
            return auth_url, state

        except ImportError as e:
            logger.error("google_auth_libs_missing", error=str(e))
            raise ImportError(
                "Install Google auth libraries: pip install google-auth google-auth-oauthlib google-auth-httplib2 google-api-python-client"
            )

    def fetch_token(self, authorization_response: str) -> bool:
        """Exchange authorization code for tokens."""
        if not self._flow:
            logger.error("google_auth_flow_not_started")
            return False

        try:
            self._flow.fetch_token(authorization_response=authorization_response)
            self._creds = self._flow.credentials
            logger.info("google_token_fetched")
            return True
        except Exception as e:
            logger.error("google_token_fetch_failed", error=str(e))
            return False

    def save_tokens(self, path: str = "~/.open-agent/google_tokens.json") -> bool:
        """Save tokens to file."""
        if not self._creds:
            return False

        try:
            token_path = Path(path).expanduser()
            token_path.parent.mkdir(parents=True, exist_ok=True)

            tokens = {
                "token": self._creds.token,
                "refresh_token": self._creds.refresh_token,
                "token_uri": self._creds.token_uri,
                "client_id": self._creds.client_id,
                "client_secret": self._creds.client_secret,
                "scopes": self._creds.scopes,
                "expiry": self._creds.expiry.timestamp() if self._creds.expiry else 0,
            }

            token_path.write_text(json.dumps(tokens, indent=2))
            logger.info("google_tokens_saved", path=str(token_path))
            return True

        except Exception as e:
            logger.error("google_tokens_save_failed", error=str(e))
            return False

    def load_tokens(self, path: str = "~/.open-agent/google_tokens.json") -> bool:
        """Load tokens from file."""
        try:
            token_path = Path(path).expanduser()
            if not token_path.exists():
                logger.warning("google_tokens_file_not_found")
                return False

            data = json.loads(token_path.read_text())
            
            from google.oauth2.credentials import Credentials
            from google.auth.transport.requests import Request

            self._creds = Credentials(
                token=data.get("token"),
                refresh_token=data.get("refresh_token"),
                token_uri=data.get("token_uri"),
                client_id=data.get("client_id"),
                client_secret=data.get("client_secret"),
                scopes=data.get("scopes"),
            )

            # Check if expired and refresh
            if self._creds.expired and self._creds.refresh_token:
                self._creds.refresh(Request())
                self.save_tokens(path)

            logger.info("google_tokens_loaded")
            return True

        except Exception as e:
            logger.error("google_tokens_load_failed", error=str(e))
            return False

    @property
    def credentials(self):
        """Get credentials."""
        return self._creds

    def is_valid(self) -> bool:
        """Check if credentials are valid."""
        return self._creds is not None and self._creds.valid


class GoogleServices:
    """Google Workspace services - matches OpenClaw's google tool."""

    def __init__(self, oauth_flow: GoogleOAuthFlow):
        self.oauth = oauth_flow
        self._services: Dict[str, Any] = {}

    def _get_service(self, name: str, version: str, build_fn: Callable):
        """Get or create a service."""
        key = f"{name}_{version}"
        if key not in self._services:
            self._services[key] = build_fn(self.oauth.credentials)
        return self._services[key]

    @property
    def gmail(self):
        """Get Gmail service."""
        from googleapiclient.discovery import build
        return self._get_service("gmail", "v1", lambda creds: build("gmail", "v1", credentials=creds, static_discovery=False))

    @property
    def calendar(self):
        """Get Calendar service."""
        from googleapiclient.discovery import build
        return self._get_service("calendar", "v3", lambda creds: build("calendar", "v3", credentials=creds, static_discovery=False))

    @property
    def drive(self):
        """Get Drive service."""
        from googleapiclient.discovery import build
        return self._get_service("drive", "v3", lambda creds: build("drive", "v3", credentials=creds, static_discovery=False))

    @property
    def sheets(self):
        """Get Sheets service."""
        from googleapiclient.discovery import build
        return self._get_service("sheets", "v4", lambda creds: build("sheets", "v4", credentials=creds, static_discovery=False))

    @property
    def docs(self):
        """Get Docs service."""
        from googleapiclient.discovery import build
        return self._get_service("docs", "v1", lambda creds: build("docs", "v1", credentials=creds, static_discovery=False))

    @property
    def searchconsole(self):
        """Get Search Console service."""
        from googleapiclient.discovery import build
        return self._get_service("searchconsole", "v1", lambda creds: build("searchconsole", "v1", credentials=creds, static_discovery=False))


class GmailTool:
    """Gmail tool - matches OpenClaw's gmail tool."""

    def __init__(self, services: GoogleServices):
        self.services = services

    async def list_emails(self, query: str = "", max_results: int = 10) -> List[Dict]:
        """List emails matching query."""
        try:
            results = (
                self.services.gmail.users()
                .messages()
                .list(userId="me", q=query, maxResults=max_results)
                .execute()
            )

            messages = []
            for msg in results.get("messages", []):
                msg_data = (
                    self.services.gmail.users()
                    .messages()
                    .get(userId="me", id=msg["id"], format="metadata")
                    .execute()
                )

                headers = msg_data.get("payload", {}).get("headers", [])
                
                def get_header(name):
                    for h in headers:
                        if h["name"].lower() == name.lower():
                            return h["value"]
                    return ""

                messages.append({
                    "id": msg["id"],
                    "from": get_header("From"),
                    "to": get_header("To"),
                    "subject": get_header("Subject"),
                    "date": get_header("Date"),
                    "snippet": msg_data.get("snippet", ""),
                })

            return messages

        except Exception as e:
            logger.error("gmail_list_failed", error=str(e))
            return []

    async def send_email(self, to: str, subject: str, body: str) -> Dict:
        """Send an email."""
        try:
            import base64
            from email.mime.text import MIMEText
            from email.mime.multipart import MIMEMultipart
            from email.mime.multipart import MIMEBase
            import email.encoders

            message = MIMEMultipart()
            message["to"] = to
            message["subject"] = subject
            message.attach(MIMEText(body, "plain"))

            encoded = base64.urlsafe_b64encode(message.as_bytes()).decode()
            
            result = (
                self.services.gmail.users()
                .messages()
                .send(userId="me", body={"raw": encoded})
                .execute()
            )

            return {"success": True, "message_id": result["id"]}

        except Exception as e:
            logger.error("gmail_send_failed", error=str(e))
            return {"success": False, "error": str(e)}

    async def get_email(self, message_id: str) -> Optional[Dict]:
        """Get full email content."""
        try:
            msg = (
                self.services.gmail.users()
                .messages()
                .get(userId="me", id=message_id, format="full")
                .execute()
            )

            headers = msg.get("payload", {}).get("headers", [])
            
            def get_header(name):
                for h in headers:
                    if h["name"].lower() == name.lower():
                        return h["value"]
                return ""

            # Get body
            body = ""
            parts = msg.get("payload", {}).get("parts", [])
            for part in parts:
                if part.get("mimeType") == "text/plain":
                    data = part.get("body", {}).get("data", "")
                    if data:
                        import base64
                        body = base64.urlsafe_b64decode(data).decode()

            return {
                "id": msg["id"],
                "from": get_header("From"),
                "to": get_header("To"),
                "subject": get_header("Subject"),
                "date": get_header("Date"),
                "body": body,
                "snippet": msg.get("snippet", ""),
            }

        except Exception as e:
            logger.error("gmail_get_failed", error=str(e))
            return None


class CalendarTool:
    """Calendar tool - matches OpenClaw's google calendar tool."""

    def __init__(self, services: GoogleServices):
        self.services = services

    async def list_events(
        self,
        days: int = 7,
        max_results: int = 50,
    ) -> List[Dict]:
        """List upcoming events."""
        try:
            from datetime import datetime, timedelta

            now = datetime.utcnow()
            time_min = now.isoformat() + "Z"
            time_max = (now + timedelta(days=days)).isoformat() + "Z"

            events = (
                self.services.calendar.events()
                .list(
                    calendarId="primary",
                    timeMin=time_min,
                    timeMax=time_max,
                    maxResults=max_results,
                    singleEvents=True,
                    orderBy="startTime",
                )
                .execute()
            )

            return [
                {
                    "id": e.get("id"),
                    "summary": e.get("summary", "No title"),
                    "start": e.get("start", {}).get("dateTime", e.get("start", {}).get("date")),
                    "end": e.get("end", {}).get("dateTime", e.get("end", {}).get("date")),
                    "location": e.get("location", ""),
                    "description": e.get("description", ""),
                    "attendees": [a.get("email") for a in e.get("attendees", [])],
                }
                for e in events.get("items", [])
            ]

        except Exception as e:
            logger.error("calendar_list_failed", error=str(e))
            return []

    async def create_event(
        self,
        summary: str,
        start: str,
        end: str,
        description: str = "",
        location: str = "",
        attendees: Optional[List[str]] = None,
    ) -> Dict:
        """Create a calendar event."""
        try:
            event = {
                "summary": summary,
                "start": {"dateTime": start, "timeZone": "UTC"},
                "end": {"dateTime": end, "timeZone": "UTC"},
                "description": description,
                "location": location,
            }

            if attendees:
                event["attendees"] = [{"email": a} for a in attendees]

            result = (
                self.services.calendar.events()
                .insert(calendarId="primary", body=event)
                .execute()
            )

            return {"success": True, "event_id": result["id"], "link": result.get("htmlLink")}

        except Exception as e:
            logger.error("calendar_create_failed", error=str(e))
            return {"success": False, "error": str(e)}


class DriveTool:
    """Drive tool - matches OpenClaw's google drive tool."""

    def __init__(self, services: GoogleServices):
        self.services = services

    async def list_files(self, folder_id: Optional[str] = None, max_results: int = 50) -> List[Dict]:
        """List files in Drive."""
        try:
            query = f"'{folder_id or 'root'}' in parents and trashed = false"
            
            results = (
                self.services.drive.files()
                .list(
                    q=query,
                    pageSize=max_results,
                    fields="files(id, name, mimeType, modifiedTime, webViewLink, shared)",
                )
                .execute()
            )

            return [
                {
                    "id": f.get("id"),
                    "name": f.get("name"),
                    "type": f.get("mimeType"),
                    "modified": f.get("modifiedTime"),
                    "link": f.get("webViewLink"),
                    "shared": f.get("shared"),
                }
                for f in results.get("files", [])
            ]

        except Exception as e:
            logger.error("drive_list_failed", error=str(e))
            return []

    async def search_files(self, query: str, max_results: int = 50) -> List[Dict]:
        """Search files by name."""
        try:
            results = (
                self.services.drive.files()
                .list(
                    q=f"name contains '{query}' and trashed = false",
                    pageSize=max_results,
                    fields="files(id, name, mimeType, modifiedTime)",
                )
                .execute()
            )

            return [
                {
                    "id": f.get("id"),
                    "name": f.get("name"),
                    "type": f.get("mimeType"),
                    "modified": f.get("modifiedTime"),
                }
                for f in results.get("files", [])
            ]

        except Exception as e:
            logger.error("drive_search_failed", error=str(e))
            return []

    async def read_file(self, file_id: str) -> Optional[str]:
        """Read file content."""
        try:
            result = (
                self.services.drive.files()
                .get_media(fileId=file_id)
                .execute()
            )
            return result.decode() if isinstance(result, bytes) else result

        except Exception as e:
            logger.error("drive_read_failed", error=str(e))
            return None

    async def create_file(self, name: str, content: str, folder_id: Optional[str] = None) -> Dict:
        """Create a new file."""
        try:
            import base64

            file_metadata = {"name": name}
            if folder_id:
                file_metadata["parents"] = [folder_id]

            media = type("Media", (), {
                "mimetype": "text/plain",
                "body": content,
            })()

            result = (
                self.services.drive.files()
                .create(body=file_metadata, media_body=media, fields="id, webViewLink")
                .execute()
            )

            return {"success": True, "file_id": result["id"], "link": result.get("webViewLink")}

        except Exception as e:
            logger.error("drive_create_failed", error=str(e))
            return {"success": False, "error": str(e)}


class GoogleWorkspace:
    """Full Google Workspace integration - matches OpenClaw."""

    def __init__(self, oauth_flow: Optional[GoogleOAuthFlow] = None):
        self.oauth = oauth_flow or GoogleOAuthFlow(GoogleOAuthConfig())
        self.services = GoogleServices(self.oauth)
        
        self.gmail = GmailTool(self.services)
        self.calendar = CalendarTool(self.services)
        self.drive = DriveTool(self.services)

    def get_auth_url(self) -> tuple[str, str]:
        """Get OAuth authorization URL."""
        return self.oauth.get_authorization_url()

    def complete_auth(self, authorization_response: str) -> bool:
        """Complete OAuth flow."""
        if self.oauth.fetch_token(authorization_response):
            return self.oauth.save_tokens()
        return False

    def load_from_file(self, path: str = "~/.open-agent/google_tokens.json") -> bool:
        """Load credentials from file."""
        return self.oauth.load_tokens(path)

    def save_to_file(self, path: str = "~/.open-agent/google_tokens.json") -> bool:
        """Save credentials to file."""
        return self.oauth.save_tokens(path)

    @property
    def is_authenticated(self) -> bool:
        """Check if authenticated."""
        return self.oauth.is_valid()


def create_google_workspace(
    client_id: Optional[str] = None,
    client_secret: Optional[str] = None,
    credentials_file: Optional[str] = None,
) -> GoogleWorkspace:
    """Create Google Workspace instance.
    
    Args:
        client_id: OAuth client ID
        client_secret: OAuth client secret
        credentials_file: Path to credentials.json file
    
    Returns:
        GoogleWorkspace instance
    """
    oauth_flow = None

    if credentials_file:
        # Load from credentials file
        try:
            import google_auth_oauthlib.flow
            oauth_flow = google_auth_oauthlib.flow.Flow.from_client_secrets_file(
                credentials_file,
                scopes=[
                    "https://www.googleapis.com/auth/gmail.readonly",
                    "https://www.googleapis.com/auth/gmail.send",
                    "https://www.googleapis.com/auth/calendar.readonly",
                    "https://www.googleapis.com/auth/calendar.events",
                    "https://www.googleapis.com/auth/drive.readonly",
                    "https://www.googleapis.com/auth/drive.file",
                ],
            )
            # Create wrapper
            config = GoogleOAuthConfig(
                client_id=client_id or "",
                client_secret=client_secret or "",
            )
            oauth_flow = GoogleOAuthFlow(config)
        except Exception as e:
            logger.error("google_credentials_load_failed", error=str(e))

    if not oauth_flow:
        # Use provided or env credentials
        config = GoogleOAuthConfig(
            client_id=client_id or os.environ.get("GOOGLE_CLIENT_ID", ""),
            client_secret=client_secret or os.environ.get("GOOGLE_CLIENT_SECRET", ""),
        )
        oauth_flow = GoogleOAuthFlow(config)

    workspace = GoogleWorkspace(oauth_flow)
    
    # Try to load from file
    workspace.load_from_file()
    
    return workspace



"""Agent runtime - Core agent execution engine."""

from typing import Optional, Dict, Any, List
from dataclasses import dataclass, field
import time
import structlog

from open_agent.agents.tools.base import BaseTool, ToolResult, ToolCall
from open_agent.agents.tools.shell import ShellTool
from open_agent.agents.tools.filesystem import FileTools
from open_agent.agents.tools.todo import TodoTool
from open_agent.agents.skills.registry import Skill
from open_agent.agents.sandbox.backend import OpenShellBackend
from open_agent.providers.nemotron import create_nemotron_model

logger = structlog.get_logger(__name__)


@dataclass
class AgentContext:
    """Context for agent execution."""
    session_id: str
    thread_id: Optional[str] = None
    user_id: Optional[str] = None
    channel: Optional[str] = None
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class Message:
    """A message in agent conversation."""
    role: str
    content: str
    name: Optional[str] = None
    tool_call_id: Optional[str] = None
    timestamp: float = field(default_factory=time.time)


@dataclass
class SoulConfig:
    """OpenClaw-style SOUL.md configuration for an agent."""
    name: str
    emoji: str = ""
    team: Optional[str] = None
    responsibilities: List[str] = field(default_factory=list)
    reporting_to: Optional[str] = None
    work_style: List[str] = field(default_factory=list)
    constraints: List[str] = field(default_factory=list)
    custom_instructions: str = ""


class DeepAgent:
    """Deep Agent powered by LangChain, Nemotron, and OpenShell.

    This is the core agent harness that combines:
    - NVIDIA Nemotron for reasoning
    - LangGraph for agentic loop
    - OpenShell for secure execution
    - Memory and skills for persistence
    """

    def __init__(
        self,
        name: str = "deep-agent",
        model: Optional[str] = None,
        system_prompt: Optional[str] = None,
        tools: Optional[List[BaseTool]] = None,
        skills: Optional[List[Skill]] = None,
        soul: Optional[SoulConfig] = None,
        openshell_backend: Optional[OpenShellBackend] = None,
        memory_enabled: bool = True,
        max_iterations: int = 50,
    ):
        self.name = name
        self.soul = soul
        self.tools = tools or []
        self.skills = skills or []
        self.openshell_backend = openshell_backend
        self.memory_enabled = memory_enabled
        self.max_iterations = max_iterations
        self._base_system_prompt = system_prompt
        self.model_name = model or "nvidia/nemotron-3-super-120b-a12b"

        self.nemotron = create_nemotron_model(
            model_name=self.model_name,
            temperature=0.7,
        )

        self.sessions: Dict[str, List[Message]] = {}
        self.current_todos: Dict[str, List[Dict[str, Any]]] = {}

        logger.info("deep_agent_initialized", name=name, model=model)

    def _build_system_prompt(self) -> str:
        """Build the system prompt from soul config."""
        parts = []

        if self.soul:
            parts.append(f"# {self.soul.emoji} {self.soul.name}\n")
            if self.soul.team:
                parts.append(f"Team: {self.soul.team}\n")
            if self.soul.responsibilities:
                parts.append("## Responsibilities\n")
                for resp in self.soul.responsibilities:
                    parts.append(f"- {resp}\n")
            if self.soul.constraints:
                parts.append("## Constraints\n")
                for constraint in self.soul.constraints:
                    parts.append(f"- {constraint}\n")

        parts.append("## Instructions\n")
        parts.append(self._base_system_prompt or "You are a helpful AI assistant.")

        if self.tools:
            parts.append("\n## Available Tools\n")
            for tool in self.tools:
                parts.append(f"- **{tool.name}**: {tool.description}\n")

        return "".join(parts)

    async def initialize(self) -> None:
        """Initialize the agent."""
        if self.openshell_backend:
            self.openshell_backend.initialize()

        for skill in self.skills:
            logger.info("skill_loaded", skill_name=skill.name)

        logger.info("agent_ready", name=self.name)

    async def process_message(
        self,
        message: str,
        context: Optional[AgentContext] = None,
    ) -> str:
        """Process a user message and return a response."""
        session_id = context.session_id if context else "default"
        thread_id = context.thread_id if context else "main"

        if session_id not in self.sessions:
            self.sessions[session_id] = []

        self.sessions[session_id].append(Message(role="user", content=message))

        from langchain_core.messages import HumanMessage, SystemMessage, AIMessage

        messages = [SystemMessage(content=self._build_system_prompt())]

        for msg in self.sessions[session_id][-20:]:
            if msg.role == "user":
                messages.append(HumanMessage(content=msg.content))
            elif msg.role == "assistant":
                if isinstance(msg.content, str):
                    messages.append(AIMessage(content=msg.content))

        try:
            response = await self.nemotron.ainvoke(messages)
            response_content = response.content if hasattr(response, 'content') else str(response)
        except Exception as e:
            logger.error("model_invoke_error", error=str(e))
            response_content = f"I encountered an error: {str(e)}"

        self.sessions[session_id].append(Message(role="assistant", content=response_content))

        return response_content

    async def spawn_subagent(
        self,
        name: str,
        task: str,
        parent_session: Optional[str] = None,
    ) -> "DeepAgent":
        """Spawn a subagent for delegated tasks."""
        subagent = DeepAgent(
            name=name,
            system_prompt=f"You are {name}, a specialized subagent. Your task: {task}",
            tools=self.tools,
            openshell_backend=self.openshell_backend,
            memory_enabled=False,
        )
        await subagent.initialize()
        return subagent

    def get_memory(self, session_id: str) -> List[Message]:
        """Get session memory."""
        return self.sessions.get(session_id, [])

    async def reset(self, session_id: str) -> None:
        """Reset agent session."""
        if session_id in self.sessions:
            self.sessions[session_id] = []
        if session_id in self.current_todos:
            del self.current_todos[session_id]
        logger.info("session_reset", session_id=session_id)

    def load_skill(self, skill: Skill) -> None:
        """Load a skill into the agent."""
        self.skills.append(skill)
        logger.info("skill_loaded", skill_name=skill.name)


def create_deep_agent(
    name: str = "deep-agent",
    model: str = "nvidia/nemotron-3-super-120b-a12b",
    system_prompt: Optional[str] = None,
    tools: Optional[List[BaseTool]] = None,
    skills: Optional[List[Skill]] = None,
    soul: Optional[SoulConfig] = None,
    use_openshell: bool = True,
    memory_enabled: bool = True,
) -> DeepAgent:
    """Create a configured deep agent."""
    openshell_backend = None
    if use_openshell:
        try:
            openshell_backend = OpenShellBackend()
        except Exception as e:
            logger.warning("openshell_not_available", error=str(e))

    agent_tools = tools or []
    if use_openshell and openshell_backend:
        agent_tools.append(ShellTool(openshell_backend))
        agent_tools.append(FileTools(openshell_backend))

    agent_tools.append(TodoTool())

    return DeepAgent(
        name=name,
        model=model,
        system_prompt=system_prompt,
        tools=agent_tools,
        skills=skills,
        soul=soul,
        openshell_backend=openshell_backend,
        memory_enabled=memory_enabled,
    )



"""Core types and interfaces for OpenAgent."""

from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Union, Optional
from pydantic import BaseModel, Field
import time


class PermissionLevel(str, Enum):
    """Permission levels for tools."""
    READ = "read"
    WRITE = "write"
    EXEC = "exec"
    ADMIN = "admin"


class ToolCall(BaseModel):
    """Represents a tool call made by an agent."""
    id: str = Field(default_factory=lambda: f"call_{id(object())}")
    name: str
    arguments: dict[str, Any] = Field(default_factory=dict)


class ToolResult(BaseModel):
    """Result of a tool execution."""
    tool_call_id: str
    success: bool
    output: Optional[str] = None
    error: Optional[str] = None


class AgentConfig(BaseModel):
    """Configuration for an agent."""
    id: str
    name: str
    description: Optional[str] = None
    model: str = "nvidia/nemotron"
    system_prompt: Optional[str] = None
    tools: list[str] = Field(default_factory=list)
    max_iterations: int = 10
    timeout: int = 300
    memory_enabled: bool = True
    permission_level: PermissionLevel = PermissionLevel.EXEC


class AgentContext(BaseModel):
    """Context passed to an agent during execution."""
    session_id: str
    user_id: Optional[str] = None
    channel: Optional[str] = None
    metadata: dict[str, Any] = Field(default_factory=dict)


class Message(BaseModel):
    """A message in a conversation."""
    role: str  # user, assistant, system, tool
    content: str
    name: Optional[str] = None
    tool_call_id: Optional[str] = None


class SessionState(BaseModel):
    """State of an agent session."""
    session_id: str
    agent_id: str
    messages: list[Message] = Field(default_factory=list)
    created_at: float = Field(default_factory=lambda: time.time())
    last_activity: float = Field(default_factory=lambda: time.time())
    metadata: dict[str, Any] = Field(default_factory=dict)


class BaseAgent(ABC):
    """Abstract base class for agents."""

    @abstractmethod
    async def initialize(self) -> None:
        """Initialize the agent."""
        pass

    @abstractmethod
    async def process_message(self, message: str, context: Optional[AgentContext] = None) -> str:
        """Process a user message and return a response."""
        pass

    @abstractmethod
    async def get_memory(self) -> list[Message]:
        """Get the agent's memory/context."""
        pass

    @abstractmethod
    async def reset(self) -> None:
        """Reset the agent's state."""
        pass



"""NVIDIA Nemotron model integration via LangChain NVIDIA AI Endpoints.

This module provides integration with NVIDIA Nemotron models using the
official langchain-nvidia-ai-endpoints package.
"""

from typing import Optional, Dict, Any, List, Union, Callable
from functools import lru_cache

from langchain_core.language_models import BaseChatModel
from langchain_core.messages import BaseMessage
from langchain_core.outputs import ChatResult

from open_agent.config.settings import get_config


def create_nemotron_model(
    model_name: Optional[str] = None,
    temperature: float = 0.7,
    max_tokens: int = 4096,
    api_key: Optional[str] = None,
    base_url: Optional[str] = None,
) -> BaseChatModel:
    """Create a configured Nemotron chat model via LangChain.

    Uses langchain-nvidia-ai-endpoints for official NVIDIA integration.

    Example:
        >>> from open_agent.models.nemotron import create_nemotron_model
        >>>
        >>> model = create_nemotron_model(
        ...     model_name="nvidia/nemotron-3-super-120b-a12b",
        ...     temperature=0.7
        ... )
    """
    try:
        from langchain_nvidia_ai_endpoints import ChatNVIDIA

        config = get_config()
        model_id = model_name or config.nvidia.model

        # Check if using hosted API or self-hosted NIM
        if base_url:
            llm = ChatNVIDIA(
                model=model_id,
                base_url=base_url,
                temperature=temperature,
                max_tokens=max_tokens,
            )
        else:
            llm = ChatNVIDIA(
                model=model_id,
                api_key=api_key or config.nvidia.api_key,
                temperature=temperature,
                max_tokens=max_tokens,
            )

        return llm

    except ImportError:
        raise ImportError(
            "langchain-nvidia-ai-endpoints is required for Nemotron integration. "
            "Install with: pip install langchain-nvidia-ai-endpoints"
        )


@lru_cache
def get_nemotron_model(
    model_name: Optional[str] = None,
    temperature: float = 0.7,
) -> BaseChatModel:
    """Get a cached Nemotron model instance."""
    return create_nemotron_model(
        model_name=model_name,
        temperature=temperature,
    )


def get_available_nemotron_models() -> List[str]:
    """Get list of available Nemotron models on NVIDIA API."""
    return [
        "nvidia/nemotron-3-super-120b-a12b",
        "nvidia/nemotron-3-ultra-405b",
        "nvidia/nemotron-3-nano-8b",
        "nvidia/Nemotron-Nano-3-30B-A3B-FP8",
        "nvidia/Nemotron-Nano-3-30B-A3B-BF16",
    ]


class NemotronModel:
    """High-level wrapper for Nemotron model with additional features.

    This class provides a simpler interface for common use cases while
    maintaining access to the underlying LangChain model.
    """

    def __init__(
        self,
        model_name: Optional[str] = None,
        temperature: float = 0.7,
        max_tokens: int = 4096,
        api_key: Optional[str] = None,
    ):
        self.model = create_nemotron_model(
            model_name=model_name,
            temperature=temperature,
            max_tokens=max_tokens,
            api_key=api_key,
        )
        self.model_name = model_name or get_config().nvidia.model

    def invoke(self, prompt: str) -> str:
        """Simple invoke call."""
        from langchain_core.messages import HumanMessage
        result = self.model.invoke([HumanMessage(content=prompt)])
        return result.content

    def stream(self, prompt: str):
        """Streaming invoke call."""
        from langchain_core.messages import HumanMessage
        return self.model.stream([HumanMessage(content=prompt)])

    def with_tools(self, tools: List[Any]) -> "NemotronModel":
        """Bind tools to the model for tool-calling support."""
        self.model = self.model.bind_tools(tools)
        return self



"""Text-to-Speech (TTS) engine for voice output."""

from typing import Optional, Dict, Any, List
from dataclasses import dataclass
from enum import Enum
import structlog
import base64
import io

logger = structlog.get_logger(__name__)


class TTSEngine(str, Enum):
    """Available TTS engines."""
    GOOGLE = "google"
    AWS_POLLY = "aws_polly"
    AZURE = "azure"
    COQUI = "coqui"
    ESPEAKE = "espeak"  # Offline fallback


@dataclass
class VoiceConfig:
    """Voice configuration."""
    engine: TTSEngine = TTSEngine.ESPEAKE
    voice_id: str = "default"
    language: str = "en-US"
    rate: float = 1.0  # Speech rate multiplier
    pitch: float = 1.0  # Pitch multiplier
    volume: float = 1.0  # Volume (0-1)
    output_format: str = "mp3"  # mp3, wav, ogg


@dataclass
class SpeechResult:
    """Result of TTS generation."""
    audio_data: bytes
    duration_seconds: float
    engine: str
    format: str


class TextToSpeech:
    """Text-to-Speech engine for converting text to audio.
    
    Similar to OpenClaw's voice system for generating
    spoken responses from the agent.
    
    Example:
        >>> tts = TextToSpeech(engine=TTSEngine.ESPEAKE)
        >>> result = await tts.speak("Hello, how can I help you?")
        >>> print(f"Audio duration: {result.duration_seconds}s")
    """

    def __init__(self, config: Optional[VoiceConfig] = None):
        self.config = config or VoiceConfig()
        self._initialized = False

    def initialize(self) -> None:
        """Initialize the TTS engine."""
        if self._initialized:
            return
        
        logger.info("tts_initialized", engine=self.config.engine)
        self._initialized = True

    async def speak(self, text: str) -> SpeechResult:
        """Convert text to speech.
        
        Args:
            text: Text to convert to speech
            
        Returns:
            SpeechResult with audio data
        """
        if not self._initialized:
            self.initialize()
        
        logger.debug("tts_speaking", text_length=len(text), engine=self.config.engine)
        
        if self.config.engine == TTSEngine.ESPEAKE:
            return await self._espeak_tts(text)
        elif self.config.engine == TTSEngine.GOOGLE:
            return await self._google_tts(text)
        elif self.config.engine == TTSEngine.AWS_POLLY:
            return await self._aws_polly_tts(text)
        else:
            return await self._espeak_tts(text)

    async def _espeak_tts(self, text: str) -> SpeechResult:
        """Use espeak for TTS (offline)."""
        try:
            import subprocess
            
            cmd = [
                "espeak",
                "-w", "/tmp/tts_output.wav",
                "-s", str(int(self.config.rate * 100)),
                text
            ]
            
            result = subprocess.run(
                cmd,
                capture_output=True,
                text=True,
                timeout=30,
            )
            
            if result.returncode == 0:
                with open("/tmp/tts_output.wav", "rb") as f:
                    audio_data = f.read()
                
                return SpeechResult(
                    audio_data=audio_data,
                    duration_seconds=len(audio_data) / 16000,
                    engine="espeak",
                    format="wav",
                )
        except FileNotFoundError:
            logger.warning("espeak_not_installed")
        except Exception as e:
            logger.error("espeak_error", error=str(e))
        
        return SpeechResult(
            audio_data=b"",
            duration_seconds=0.0,
            engine="espeak",
            format="wav",
        )

    async def _google_tts(self, text: str) -> SpeechResult:
        """Use Google Cloud TTS."""
        try:
            from google.cloud import texttospeech
            
            client = texttospeech.TextToSpeechClient()
            
            synthesis_input = texttospeech.SynthesisInput(text=text)
            
            voice = texttospeech.VoiceSelectionParams(
                language_code=self.config.language,
                name=self.config.voice_id,
            )
            
            audio_config = texttospeech.AudioConfig(
                audio_encoding=texttospeech.AudioEncoding.MP3,
                speaking_rate=self.config.rate,
                pitch=self.config.pitch - 1.0,  # Google uses -20 to +20
            )
            
            response = client.synthesize_speech(
                input=synthesis_input,
                voice=voice,
                audio_config=audio_config,
            )
            
            return SpeechResult(
                audio_data=response.audio_content,
                duration_seconds=len(response.audio_content) / 16000,
                engine="google",
                format="mp3",
            )
        except ImportError:
            logger.warning("google_cloud_not_installed")
        except Exception as e:
            logger.error("google_tts_error", error=str(e))
        
        return await self._espeak_tts(text)

    async def _aws_polly_tts(self, text: str) -> SpeechResult:
        """Use AWS Polly TTS."""
        try:
            import boto3
            
            polly = boto3.client("polly")
            
            response = polly.synthesize_speech(
                Text=text,
                OutputFormat=self.config.output_format.upper(),
                VoiceId=self.config.voice_id,
                Engine="neural" if "neural" in self.config.voice_id else "standard",
            )
            
            audio_stream = response["AudioStream"].read()
            
            return SpeechResult(
                audio_data=audio_stream,
                duration_seconds=len(audio_stream) / 16000,
                engine="aws_polly",
                format=self.config.output_format,
            )
        except ImportError:
            logger.warning("boto3_not_installed")
        except Exception as e:
            logger.error("aws_polly_error", error=str(e))
        
        return await self._espeak_tts(text)

    def get_available_voices(self) -> List[Dict[str, str]]:
        """Get list of available voices for the current engine."""
        default_voices = [
            {"id": "default", "name": "Default", "language": "en-US"},
            {"id": "female", "name": "Female", "language": "en-US"},
            {"id": "male", "name": "Male", "language": "en-US"},
        ]
        
        if self.config.engine == TTSEngine.GOOGLE:
            return [
                {"id": "en-US-Neural2-F", "name": "Neural2 Female", "language": "en-US"},
                {"id": "en-US-Neural2-J", "name": "Neural2 Male", "language": "en-US"},
            ]
        elif self.config.engine == TTSEngine.AWS_POLLY:
            return [
                {"id": "Joanna", "name": "Joanna (Female)", "language": "en-US"},
                {"id": "Matthew", "name": "Matthew (Male)", "language": "en-US"},
            ]
        
        return default_voices


def create_tts_engine(
    engine: str = "espeak",
    voice_id: str = "default",
    language: str = "en-US",
) -> TextToSpeech:
    """Create a TTS engine.
    
    Example:
        >>> tts = create_tts_engine(engine="espeak")
    """
    return TextToSpeech(
        config=VoiceConfig(
            engine=TTSEngine(engine),
            voice_id=voice_id,
            language=language,
        )
    )



"""OpenClaw-style agent runtime.

This module provides the core agent runtime with:
- Tool execution system
- Skill-based extensibility
- Sandbox execution
- Auth profiles for model access
"""

from open_agent.agents.tools.base import BaseTool, ToolResult, ToolCall
from open_agent.agents.tools.shell import ShellTool
from open_agent.agents.tools.filesystem import FileTools
from open_agent.agents.tools.todo import TodoTool
from open_agent.agents.skills.registry import Skill, SkillRegistry, SkillInvocation
from open_agent.agents.sandbox.backend import (
    OpenShellBackend,
    SandboxSession,
    SandboxType,
    create_openshell_backend,
)
from open_agent.agents.auth.profiles import AuthProfile, AuthProfileStore
from open_agent.agents.runtime import DeepAgent, AgentContext, SoulConfig, Message

__all__ = [
    # Tools
    "BaseTool",
    "ToolResult", 
    "ToolCall",
    "ShellTool",
    "FileTools",
    "TodoTool",
    # Skills
    "Skill",
    "SkillRegistry",
    "SkillInvocation",
    # Sandbox
    "OpenShellBackend",
    "SandboxSession",
    "SandboxType",
    "create_openshell_backend",
    # Auth
    "AuthProfile",
    "AuthProfileStore",
    # Runtime
    "DeepAgent",
    "AgentContext",
    "SoulConfig",
    "Message",
]



"""Gateway - Central routing and session management for OpenAgent.

Inspired by OpenClaw's Gateway architecture:
- Channel adapters for multi-platform support
- Session management with persistent state
- Agent registry and routing
- WebSocket/HTTP endpoints for external access
"""

from typing import Optional, Dict, Any, List, Callable
from dataclasses import dataclass, field
from enum import Enum
import asyncio
import time
import structlog

from open_agent.orchestration.openclaw import DeepAgent, AgentContext, Message

logger = structlog.get_logger(__name__)


class Channel(str, Enum):
    """Supported communication channels."""
    TERMINAL = "terminal"
    HTTP = "http"
    WEBSOCKET = "websocket"
    SLACK = "slack"
    DISCORD = "discord"
    TELEGRAM = "telegram"


@dataclass
class Binding:
    """Message routing binding."""
    channel: Channel
    account_id: str
    agent_id: str
    priority: int = 0


@dataclass
class Session:
    """An active conversation session."""
    session_id: str
    agent_id: str
    user_id: Optional[str] = None
    channel: Optional[Channel] = None
    created_at: float = field(default_factory=time.time)
    last_activity: float = field(default_factory=time.time)
    metadata: Dict[str, Any] = field(default_factory=dict)
    message_count: int = 0


class Gateway:
    """Central gateway for message routing and agent orchestration.

    Key features:
    - Multi-agent support with agent registry
    - Session management with context persistence
    - Channel adapters for various platforms
    - Binding-based message routing
    """

    def __init__(
        self,
        name: str = "open-agent-gateway",
        enable_http: bool = False,
        enable_websocket: bool = False,
        port: int = 8080,
    ):
        self.name = name
        self.port = port
        self.enable_http = enable_http
        self.enable_websocket = enable_websocket

        self._agents: Dict[str, DeepAgent] = {}
        self._sessions: Dict[str, Session] = {}
        self._bindings: List[Binding] = []
        self._http_server = None
        self._ws_server = None

        logger.info("gateway_initialized", name=name, port=port)

    def register_agent(self, agent: DeepAgent) -> None:
        """Register an agent with the gateway."""
        self._agents[agent.name] = agent
        logger.info("agent_registered", agent_name=agent.name, agent_id=agent.name)

    def unregister_agent(self, agent_id: str) -> bool:
        """Unregister an agent."""
        if agent_id in self._agents:
            del self._agents[agent_id]
            logger.info("agent_unregistered", agent_id=agent_id)
            return True
        return False

    def get_agent(self, agent_id: str) -> Optional[DeepAgent]:
        """Get an agent by ID."""
        return self._agents.get(agent_id)

    def list_agents(self) -> List[Dict[str, str]]:
        """List all registered agents."""
        return [{"id": name, "name": agent.name} for name, agent in self._agents.items()]

    def add_binding(self, binding: Binding) -> None:
        """Add a routing binding."""
        self._bindings.append(binding)
        self._bindings.sort(key=lambda b: b.priority, reverse=True)
        logger.info("binding_added", channel=binding.channel, agent_id=binding.agent_id)

    def route_message(self, channel: str, account_id: str) -> Optional[str]:
        """Route a message to the appropriate agent based on bindings."""
        for binding in self._bindings:
            if binding.channel == channel and binding.account_id == account_id:
                return binding.agent_id
        # Fallback to first registered agent
        if self._agents:
            return list(self._agents.keys())[0]
        return None

    async def create_session(
        self,
        agent_id: str,
        user_id: Optional[str] = None,
        channel: Optional[Channel] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> Session:
        """Create a new session."""
        if agent_id not in self._agents:
            raise ValueError(f"Agent '{agent_id}' not found")

        session_id = f"session_{int(time.time() * 1000)}"
        session = Session(
            session_id=session_id,
            agent_id=agent_id,
            user_id=user_id,
            channel=channel,
            metadata=metadata or {},
        )
        self._sessions[session_id] = session
        await self._agents[agent_id].initialize()

        logger.info("session_created", session_id=session_id, agent_id=agent_id)
        return session

    async def process_message(
        self,
        session_id: str,
        content: str,
    ) -> str:
        """Process a message in a session."""
        session = self._sessions.get(session_id)
        if not session:
            raise ValueError(f"Session '{session_id}' not found")

        agent = self._agents.get(session.agent_id)
        if not agent:
            raise ValueError(f"Agent '{session.agent_id}' not found")

        context = AgentContext(
            session_id=session_id,
            user_id=session.user_id,
            channel=session.channel.value if session.channel else None,
        )

        session.last_activity = time.time()
        session.message_count += 1

        logger.debug("processing_message", session_id=session_id, message_length=len(content))

        response = await agent.process_message(content, context)

        session.last_activity = time.time()
        return response

    def get_session(self, session_id: str) -> Optional[Session]:
        """Get a session by ID."""
        return self._sessions.get(session_id)

    def list_sessions(self) -> List[Session]:
        """List all active sessions."""
        return list(self._sessions.values())

    async def close_session(self, session_id: str) -> bool:
        """Close a session."""
        if session_id in self._sessions:
            del self._sessions[session_id]
            logger.info("session_closed", session_id=session_id)
            return True
        return False

    def get_stats(self) -> Dict[str, Any]:
        """Get gateway statistics."""
        return {
            "agent_count": len(self._agents),
            "session_count": len(self._sessions),
            "binding_count": len(self._bindings),
            "total_messages": sum(s.message_count for s in self._sessions.values()),
        }

    async def start(self) -> None:
        """Start the gateway servers."""
        if self.enable_http:
            await self._start_http_server()

        if self.enable_websocket:
            await self._start_websocket_server()

        logger.info("gateway_started", port=self.port)

    async def stop(self) -> None:
        """Stop the gateway servers."""
        if self._http_server:
            self._http_server.close()
        if self._ws_server:
            self._ws_server.close()
        logger.info("gateway_stopped")

    async def _start_http_server(self) -> None:
        """Start HTTP server for REST API."""
        try:
            from fastapi import FastAPI, HTTPException
            from fastapi.responses import JSONResponse
            import uvicorn

            app = FastAPI(title="OpenAgent Gateway")

            @app.get("/health")
            async def health():
                return {"status": "healthy", "gateway": self.name}

            @app.get("/agents")
            async def list_agents():
                return {"agents": self.list_agents()}

            @app.get("/sessions")
            async def list_sessions():
                return {"sessions": [s.session_id for s in self.list_sessions()]}

            @app.post("/sessions")
            async def create_session_endpoint(request: Dict[str, Any]):
                session = await self.create_session(
                    agent_id=request.get("agent_id", "default"),
                    user_id=request.get("user_id"),
                )
                return {"session_id": session.session_id}

            @app.get("/sessions/{session_id}")
            async def get_session_endpoint(session_id: str):
                session = self.get_session(session_id)
                if not session:
                    raise HTTPException(status_code=404, detail="Session not found")
                return session.__dict__

            @app.post("/sessions/{session_id}/messages")
            async def send_message(session_id: str, request: Dict[str, Any]):
                content = request.get("content", "")
                response = await self.process_message(session_id, content)
                return {"response": response}

            @app.get("/stats")
            async def stats():
                return self.get_stats()

            self._http_server = uvicorn.Server(
                uvicorn.Config(app, host="0.0.0.0", port=self.port)
            )
            await self._http_server.serve()

        except ImportError:
            logger.warning("fastapi_not_available", message="Install fastapi and uvicorn for HTTP API")

    async def _start_websocket_server(self) -> None:
        """Start WebSocket server for real-time communication."""
        try:
            import websockets

            clients: Dict[str, Any] = {}

            async def handle_client(websocket, path):
                client_id = str(id(websocket))
                clients[client_id] = websocket
                logger.info("ws_client_connected", client_id=client_id)

                try:
                    async for message in websocket:
                        import json
                        data = json.loads(message)

                        if data.get("type") == "create_session":
                            session = await self.create_session(
                                agent_id=data.get("agent_id", "default"),
                                user_id=data.get("user_id"),
                            )
                            await websocket.send(json.dumps({
                                "type": "session_created",
                                "session_id": session.session_id,
                            }))

                        elif data.get("type") == "message":
                            session_id = data.get("session_id")
                            content = data.get("content", "")
                            response = await self.process_message(session_id, content)
                            await websocket.send(json.dumps({
                                "type": "response",
                                "content": response,
                            }))

                except Exception as e:
                    logger.error("ws_client_error", client_id=client_id, error=str(e))
                finally:
                    del clients[client_id]
                    logger.info("ws_client_disconnected", client_id=client_id)

            self._ws_server = await websockets.serve(
                handle_client,
                "0.0.0.0",
                self.port + 1,
            )

        except ImportError:
            logger.warning("websockets_not_available", message="Install websockets for WebSocket support")


def create_gateway(
    name: str = "open-agent-gateway",
    port: int = 8080,
    enable_http: bool = True,
    enable_websocket: bool = True,
) -> Gateway:
    """Create a configured gateway instance."""
    return Gateway(
        name=name,
        port=port,
        enable_http=enable_http,
        enable_websocket=enable_websocket,
    )



"""Canvas workspace - Interactive code editing environment."""

from typing import Optional, Dict, List, Any, Callable
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
import re
import structlog

logger = structlog.get_logger(__name__)


class FileType(str, Enum):
    """Supported file types."""
    PYTHON = "python"
    JAVASCRIPT = "javascript"
    TYPESCRIPT = "typescript"
    HTML = "html"
    CSS = "css"
    JSON = "json"
    MARKDOWN = "markdown"
    YAML = "yaml"
    TEXT = "text"


@dataclass
class Cursor:
    """Cursor position in a file."""
    line: int = 0
    column: int = 0


@dataclass
class Selection:
    """Text selection in a file."""
    start: Cursor
    end: Cursor
    text: str = ""


@dataclass
class CanvasFile:
    """A file in the canvas workspace."""
    path: str
    content: str
    language: FileType = FileType.TEXT
    modified: bool = False
    created_at: datetime = field(default_factory=datetime.now)
    modified_at: datetime = field(default_factory=datetime.now)
    cursor: Cursor = field(default_factory=Cursor)
    selection: Optional[Selection] = None

    def get_language_from_extension(self) -> FileType:
        """Detect language from file extension."""
        ext_map = {
            ".py": FileType.PYTHON,
            ".js": FileType.JAVASCRIPT,
            ".ts": FileType.TYPESCRIPT,
            ".html": FileType.HTML,
            ".css": FileType.CSS,
            ".json": FileType.JSON,
            ".md": FileType.MARKDOWN,
            ".yaml": FileType.YAML,
            ".yml": FileType.YAML,
        }
        
        for ext, lang in ext_map.items():
            if self.path.endswith(ext):
                return lang
        return FileType.TEXT


@dataclass
class CanvasSession:
    """A canvas editing session."""
    session_id: str
    files: Dict[str, CanvasFile] = field(default_factory=dict)
    created_at: datetime = field(default_factory=datetime.now)
    last_activity: datetime = field(default_factory=datetime.now)
    metadata: Dict[str, Any] = field(default_factory=dict)

    def add_file(self, path: str, content: str = "") -> CanvasFile:
        """Add a file to the session."""
        file = CanvasFile(path=path, content=content)
        file.language = file.get_language_from_extension()
        self.files[path] = file
        self.last_activity = datetime.now()
        return file

    def get_file(self, path: str) -> Optional[CanvasFile]:
        """Get a file by path."""
        return self.files.get(path)

    def remove_file(self, path: str) -> bool:
        """Remove a file from the session."""
        if path in self.files:
            del self.files[path]
            self.last_activity = datetime.now()
            return True
        return False

    def list_files(self) -> List[str]:
        """List all files in the session."""
        return list(self.files.keys())


class Canvas:
    """Interactive canvas workspace for code editing.
    
    Similar to OpenClaw's Canvas for providing an interactive
    coding environment for agents.
    
    Example:
        >>> canvas = Canvas()
        >>> session = canvas.create_session("my-workspace")
        >>> canvas.write_file(session.session_id, "main.py", "print('Hello')")
        >>> content = canvas.read_file(session.session_id, "main.py")
        >>> print(content)
    """

    def __init__(self, root_path: Optional[str] = None):
        self.root_path = root_path or "/tmp/canvas"
        self.sessions: Dict[str, CanvasSession] = {}
        self._hooks: Dict[str, List[Callable]] = {
            "file_created": [],
            "file_modified": [],
            "file_deleted": [],
            "session_created": [],
            "session_closed": [],
        }

    def create_session(
        self,
        session_id: str,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> CanvasSession:
        """Create a new canvas session."""
        session = CanvasSession(
            session_id=session_id,
            metadata=metadata or {},
        )
        self.sessions[session_id] = session
        self._trigger_hook("session_created", session)
        logger.info("canvas_session_created", session_id=session_id)
        return session

    def get_session(self, session_id: str) -> Optional[CanvasSession]:
        """Get a session by ID."""
        return self.sessions.get(session_id)

    def close_session(self, session_id: str) -> bool:
        """Close a canvas session."""
        if session_id in self.sessions:
            session = self.sessions[session_id]
            del self.sessions[session_id]
            self._trigger_hook("session_closed", session)
            logger.info("canvas_session_closed", session_id=session_id)
            return True
        return False

    def write_file(
        self,
        session_id: str,
        path: str,
        content: str,
    ) -> bool:
        """Write content to a file."""
        session = self.get_session(session_id)
        if not session:
            return False

        existing = session.get_file(path)
        if existing:
            existing.content = content
            existing.modified = True
            existing.modified_at = datetime.now()
        else:
            session.add_file(path, content)

        self._trigger_hook("file_modified", session, path, content)
        logger.debug("canvas_file_written", session_id=session_id, path=path)
        return True

    def read_file(self, session_id: str, path: str) -> Optional[str]:
        """Read content from a file."""
        session = self.get_session(session_id)
        if not session:
            return None

        file = session.get_file(path)
        if file:
            return file.content
        return None

    def delete_file(self, session_id: str, path: str) -> bool:
        """Delete a file from the session."""
        session = self.get_session(session_id)
        if not session:
            return False

        if session.remove_file(path):
            self._trigger_hook("file_deleted", session, path)
            return True
        return False

    def execute_file(
        self,
        session_id: str,
        path: str,
        interpreter: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Execute a file and return results.
        
        Args:
            session_id: Canvas session ID
            path: Path to the file to execute
            interpreter: Override interpreter (e.g., "python3", "node")
            
        Returns:
            Dict with stdout, stderr, returncode
        """
        import subprocess
        import tempfile
        import os

        content = self.read_file(session_id, path)
        if content is None:
            return {"success": False, "error": "File not found", "stdout": "", "stderr": ""}

        session = self.get_session(session_id)
        file = session.get_file(path)
        
        if file.language == FileType.PYTHON:
            cmd = [interpreter or "python3"]
        elif file.language == FileType.JAVASCRIPT:
            cmd = [interpreter or "node"]
        elif file.language == FileType.TYPESCRIPT:
            cmd = ["npx", "ts-node"]
        else:
            return {
                "success": False,
                "error": f"No interpreter for {file.language}",
                "stdout": "",
                "stderr": "",
            }

        with tempfile.NamedTemporaryFile(
            mode='w',
            suffix=path,
            delete=False
        ) as f:
            f.write(content)
            temp_path = f.name

        try:
            result = subprocess.run(
                cmd + [temp_path],
                capture_output=True,
                text=True,
                timeout=30,
            )
            
            return {
                "success": result.returncode == 0,
                "stdout": result.stdout,
                "stderr": result.stderr,
                "returncode": result.returncode,
            }
        except subprocess.TimeoutExpired:
            return {
                "success": False,
                "error": "Execution timed out",
                "stdout": "",
                "stderr": "",
                "returncode": -1,
            }
        except FileNotFoundError:
            return {
                "success": False,
                "error": f"{cmd[0]} not found",
                "stdout": "",
                "stderr": "",
                "returncode": -1,
            }
        finally:
            os.unlink(temp_path)

    def apply_edit(
        self,
        session_id: str,
        path: str,
        old_text: str,
        new_text: str,
    ) -> bool:
        """Apply a text edit to a file.
        
        Args:
            session_id: Canvas session ID
            path: File path
            old_text: Text to replace
            new_text: Replacement text
            
        Returns:
            True if edit was applied
        """
        content = self.read_file(session_id, path)
        if content is None:
            return False

        if old_text not in content:
            return False

        new_content = content.replace(old_text, new_text)
        return self.write_file(session_id, path, new_content)

    def insert_text(
        self,
        session_id: str,
        path: str,
        text: str,
        line: Optional[int] = None,
        column: int = 0,
    ) -> bool:
        """Insert text at a specific position.
        
        Args:
            session_id: Canvas session ID
            path: File path
            text: Text to insert
            line: Line number (0-indexed), appends if None
            column: Column position
            
        Returns:
            True if text was inserted
        """
        content = self.read_file(session_id, path)
        if content is None:
            return False

        lines = content.split('\n')
        
        if line is None:
            lines.append(text)
        elif line < len(lines):
            line_content = lines[line]
            lines[line] = line_content[:column] + text + line_content[column:]
        else:
            while len(lines) < line:
                lines.append('')
            lines.append(text)

        new_content = '\n'.join(lines)
        return self.write_file(session_id, path, new_content)

    def get_file_tree(self, session_id: str) -> Dict[str, Any]:
        """Get a tree structure of all files in the session."""
        session = self.get_session(session_id)
        if not session:
            return {}

        tree = {"type": "folder", "name": "/", "children": {}}
        
        for path in sorted(session.files.keys()):
            parts = path.strip("/").split("/")
            current = tree["children"]
            
            for i, part in enumerate(parts):
                if i == len(parts) - 1:
                    file = session.get_file(path)
                    current[part] = {
                        "type": "file",
                        "name": part,
                        "path": path,
                        "language": file.language.value if file else "text",
                        "modified": file.modified if file else False,
                    }
                else:
                    if part not in current:
                        current[part] = {"type": "folder", "name": part, "children": {}}
                    current = current[part]["children"]

        return tree

    def register_hook(self, event: str, handler: Callable) -> None:
        """Register a hook handler."""
        if event in self._hooks:
            self._hooks[event].append(handler)

    def _trigger_hook(self, event: str, *args, **kwargs) -> None:
        """Trigger all handlers for an event."""
        for handler in self._hooks.get(event, []):
            try:
                handler(*args, **kwargs)
            except Exception as e:
                logger.error("canvas_hook_error", event=event, error=str(e))

    def list_sessions(self) -> List[str]:
        """List all active sessions."""
        return list(self.sessions.keys())


def create_canvas(root_path: Optional[str] = None) -> Canvas:
    """Create a new canvas workspace.
    
    Example:
        >>> canvas = create_canvas("/tmp/my-workspace")
    """
    return Canvas(root_path=root_path)



"""Browser sandbox module - Browser automation for agents.

This module provides browser sandbox functionality similar to OpenClaw:
- Playwright-based browser automation
- Screenshot capture
- Web scraping
- Form filling
"""

from open_agent.sandbox.browser import (
    BrowserSandbox, BrowserConfig, BrowserType, BrowserAction,
    BrowserResult, create_browser_sandbox
)

__all__ = [
    "BrowserSandbox",
    "BrowserConfig",
    "BrowserType",
    "BrowserAction",
    "BrowserResult",
    "create_browser_sandbox",
]



"""Sandbox module - Secure execution environments for agents."""

from open_agent.agents.sandbox.backend import (
    OpenShellBackend,
    SandboxSession,
    SandboxType,
    create_openshell_backend,
)

__all__ = [
    "OpenShellBackend",
    "SandboxSession",
    "SandboxType",
    "create_openshell_backend",
]



"""Session manager - Manages agent sessions and conversation history."""

from typing import Optional, Dict, Any, List
from dataclasses import dataclass, field
from datetime import datetime
import asyncio
import structlog

logger = structlog.get_logger(__name__)


@dataclass
class Session:
    """An agent session."""
    session_id: str
    user_id: Optional[str] = None
    channel: Optional[str] = None
    created_at: datetime = field(default_factory=datetime.now)
    last_activity: datetime = field(default_factory=datetime.now)
    metadata: Dict[str, Any] = field(default_factory=dict)
    message_count: int = 0

    def update_activity(self) -> None:
        """Update last activity timestamp."""
        self.last_activity = datetime.now()
        self.message_count += 1


class SessionManager:
    """Manages agent sessions and conversation history.
    
    Similar to OpenClaw's sessions system for managing
    conversation contexts across channels.
    """

    def __init__(self, max_sessions: int = 1000, session_timeout: int = 3600):
        self.sessions: Dict[str, Session] = {}
        self.max_sessions = max_sessions
        self.session_timeout = session_timeout
        self._cleanup_task: Optional[asyncio.Task] = None

    async def create_session(
        self,
        session_id: str,
        user_id: Optional[str] = None,
        channel: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> Session:
        """Create a new session."""
        if session_id in self.sessions:
            return self.sessions[session_id]

        if len(self.sessions) >= self.max_sessions:
            await self._cleanup_old_sessions()

        session = Session(
            session_id=session_id,
            user_id=user_id,
            channel=channel,
            metadata=metadata or {},
        )
        self.sessions[session_id] = session
        logger.info("session_created", session_id=session_id)
        return session

    async def get_session(self, session_id: str) -> Optional[Session]:
        """Get a session by ID."""
        session = self.sessions.get(session_id)
        if session:
            session.update_activity()
        return session

    async def delete_session(self, session_id: str) -> bool:
        """Delete a session."""
        if session_id in self.sessions:
            del self.sessions[session_id]
            logger.info("session_deleted", session_id=session_id)
            return True
        return False

    async def list_sessions(
        self,
        user_id: Optional[str] = None,
        channel: Optional[str] = None,
    ) -> List[Session]:
        """List sessions, optionally filtered."""
        sessions = list(self.sessions.values())
        
        if user_id:
            sessions = [s for s in sessions if s.user_id == user_id]
        if channel:
            sessions = [s for s in sessions if s.channel == channel]
        
        return sessions

    async def _cleanup_old_sessions(self) -> None:
        """Clean up old sessions."""
        now = datetime.now()
        old_sessions = [
            sid for sid, session in self.sessions.items()
            if (now - session.last_activity).total_seconds() > self.session_timeout
        ]
        
        for session_id in old_sessions:
            await self.delete_session(session_id)
        
        if old_sessions:
            logger.info("sessions_cleaned", count=len(old_sessions))

    async def start_cleanup_task(self) -> None:
        """Start the periodic cleanup task."""
        if self._cleanup_task is not None:
            return

        async def cleanup_loop():
            while True:
                await asyncio.sleep(300)
                await self._cleanup_old_sessions()

        self._cleanup_task = asyncio.create_task(cleanup_loop())

    async def stop_cleanup_task(self) -> None:
        """Stop the periodic cleanup task."""
        if self._cleanup_task:
            self._cleanup_task.cancel()
            self._cleanup_task = None



"""Base tool classes for agent tool system."""

from typing import Optional, Dict, Any, List
from dataclasses import dataclass, field
from enum import Enum
import time


class PermissionLevel(str, Enum):
    """Permission levels for agent actions."""
    READ = "read"
    WRITE = "write"
    EXEC = "exec"
    ADMIN = "admin"


@dataclass
class ToolCall:
    """A tool call made by an agent."""
    id: str
    name: str
    arguments: Dict[str, Any]


@dataclass
class ToolResult:
    """Result of a tool execution."""
    tool_call_id: str
    success: bool
    output: Optional[str] = None
    error: Optional[str] = None
    timestamp: float = field(default_factory=time.time)
    metadata: Dict[str, Any] = field(default_factory=dict)


class BaseTool:
    """Base class for agent tools.
    
    Tools are the primary way for agents to interact with the outside world.
    They provide capabilities like shell execution, file operations, and
    integration with external services.
    
    Example:
        >>> class MyTool(BaseTool):
        ...     def __init__(self):
        ...         super().__init__(
        ...             name="my_tool",
        ...             description="Does something useful",
        ...             permission_level=PermissionLevel.EXEC,
        ...         )
        ...
        ...     async def execute(self, **kwargs) -> str:
        ...         return "Result"
    """

    def __init__(
        self,
        name: str,
        description: str,
        permission_level: PermissionLevel = PermissionLevel.EXEC,
    ):
        self.name = name
        self.description = description
        self.permission_level = permission_level

    async def execute(self, **kwargs) -> str:
        """Execute the tool with the given arguments.
        
        Args:
            **kwargs: Tool-specific arguments
            
        Returns:
            Tool execution result as a string
        """
        raise NotImplementedError

    def get_schema(self) -> Dict[str, Any]:
        """Get the JSON schema for this tool's parameters."""
        return {
            "name": self.name,
            "description": self.description,
            "parameters": {
                "type": "object",
                "properties": {},
                "required": [],
            },
        }


class ToolRegistry:
    """Registry for managing agent tools."""

    def __init__(self):
        self.tools: Dict[str, BaseTool] = {}

    def register(self, tool: BaseTool) -> None:
        """Register a tool."""
        self.tools[tool.name] = tool

    def get(self, name: str) -> Optional[BaseTool]:
        """Get a tool by name."""
        return self.tools.get(name)

    def list(self) -> List[BaseTool]:
        """List all registered tools."""
        return list(self.tools.values())

    async def execute(self, tool_name: str, **kwargs) -> ToolResult:
        """Execute a tool by name."""
        tool = self.tools.get(tool_name)
        if not tool:
            return ToolResult(
                tool_call_id=kwargs.get("tool_call_id", ""),
                success=False,
                error=f"Tool '{tool_name}' not found",
            )

        try:
            output = await tool.execute(**kwargs)
            return ToolResult(
                tool_call_id=kwargs.get("tool_call_id", ""),
                success=True,
                output=output,
            )
        except Exception as e:
            return ToolResult(
                tool_call_id=kwargs.get("tool_call_id", ""),
                success=False,
                error=str(e),
            )



"""Built-in tools for agents."""

from typing import Optional, List, Dict, Any, Type
from dataclasses import dataclass
import json
import structlog

from open_agent.agents.tools.base import BaseTool

logger = structlog.get_logger(__name__)


@dataclass
class CalculatorTool(BaseTool):
    """Calculator tool for mathematical operations."""
    
    name: str = "calculator"
    description: str = "Performs mathematical calculations. Input should be a mathematical expression."
    
    def execute(self, expression: str) -> str:
        """Execute a mathematical expression."""
        try:
            import re
            safe_dict = {
                "abs": abs, "round": round, "min": min, "max": max,
                "sum": sum, "pow": pow, "sqrt": __import__("math").sqrt,
                "sin": __import__("math").sin, "cos": __import__("math").cos,
                "tan": __import__("math").tan, "log": __import__("math").log,
                "pi": __import__("math").pi, "e": __import__("math").e,
            }
            result = eval(expression, {"__builtins__": {}}, safe_dict)
            return f"Result: {result}"
        except Exception as e:
            return f"Error: {str(e)}"


@dataclass
class SearchTool(BaseTool):
    """Web search tool using web search API."""
    
    name: str = "search"
    description: str = "Search the web for information. Input should be a search query."
    
    def execute(self, query: str) -> str:
        """Search for information."""
        try:
            from open_agent.integrations.google import search_web
            results = search_web(query, num_results=5)
            return results
        except ImportError:
            return "Search tool requires google-api-python-client. Install with: pip install google-api-python-client"
        except Exception as e:
            return f"Search error: {str(e)}"


@dataclass
class WikipediaTool(BaseTool):
    """Wikipedia search tool."""
    
    name: str = "wikipedia"
    description: str = "Search Wikipedia for information. Input should be a topic to search."
    
    def execute(self, topic: str) -> str:
        """Search Wikipedia."""
        try:
            import wikipedia
            summary = wikipedia.summary(topic, sentences=3)
            return summary
        except ImportError:
            return "Wikipedia tool requires wikipedia package. Install with: pip install wikipedia"
        except wikipedia.exceptions.DisambiguationError as e:
            return f"Multiple results found: {', '.join(e.options[:5])}"
        except Exception as e:
            return f"Wikipedia error: {str(e)}"


@dataclass
class FileReadTool(BaseTool):
    """Read files from the filesystem."""
    
    name: str = "read_file"
    description: str = "Read contents of a file. Input should be a file path."
    
    def execute(self, file_path: str) -> str:
        """Read a file."""
        try:
            with open(file_path, 'r') as f:
                return f.read()
        except FileNotFoundError:
            return f"File not found: {file_path}"
        except Exception as e:
            return f"Error reading file: {str(e)}"


@dataclass
class FileWriteTool(BaseTool):
    """Write files to the filesystem."""
    
    name: str = "write_file"
    description: str = "Write content to a file. Input should be in format: file_path|content"
    
    def execute(self, input_str: str) -> str:
        """Write to a file."""
        try:
            file_path, content = input_str.split("|", 1)
            with open(file_path, 'w') as f:
                f.write(content)
            return f"Successfully wrote to {file_path}"
        except Exception as e:
            return f"Error writing file: {str(e)}"


@dataclass  
class WebFetchTool(BaseTool):
    """Fetch content from a URL."""
    
    name: str = "fetch_url"
    description: str = "Fetch content from a URL. Input should be a URL."
    
    def execute(self, url: str) -> str:
        """Fetch URL content."""
        try:
            import httpx
            response = httpx.get(url, timeout=10)
            response.raise_for_status()
            content = response.text[:5000]
            if len(response.text) > 5000:
                content += "\n... (truncated)"
            return content
        except Exception as e:
            return f"Error fetching URL: {str(e)}"


@dataclass
class PythonREPLTool(BaseTool):
    """Execute Python code in a REPL."""
    
    name: str = "python_repl"
    description: str = "Execute Python code. Input should be Python code to execute."
    
    def execute(self, code: str) -> str:
        """Execute Python code."""
        try:
            import io
            import contextlib
            
            output = io.StringIO()
            with contextlib.redirect_stdout(output):
                exec(code, {"__builtins__": __builtins__})
            
            result = output.getvalue()
            return result if result else "Code executed successfully (no output)"
        except Exception as e:
            return f"Error: {type(e).__name__}: {str(e)}"


@dataclass
class GitHubTool(BaseTool):
    """GitHub operations tool."""
    
    name: str = "github"
    description: str = "Perform GitHub operations. Input should be a JSON command."
    
    def execute(self, command_json: str) -> str:
        """Execute GitHub command."""
        try:
            command = json.loads(command_json)
            action = command.get("action", "info")
            
            if action == "search_repos":
                from open_agent.integrations.github import search_repositories
                query = command.get("query", "")
                return search_repositories(query)
            else:
                return f"Unknown action: {action}"
        except Exception as e:
            return f"GitHub error: {str(e)}"


@dataclass
class JiraTool(BaseTool):
    """Jira integration tool."""
    
    name: str = "jira"
    description: str = "Interact with Jira. Input should be a JSON command."
    
    def execute(self, command_json: str) -> str:
        """Execute Jira command."""
        try:
            command = json.loads(command_json)
            from open_agent.integrations.jira import JiraClient
            
            jira_url = command.get("url", "")
            api_token = command.get("token", "")
            email = command.get("email", "")
            
            if command.get("action") == "create_issue":
                client = JiraClient(jira_url, email, api_token)
                issue = client.create_issue(
                    project=command.get("project"),
                    summary=command.get("summary"),
                    description=command.get("description"),
                    issue_type=command.get("issue_type", "Task")
                )
                return f"Created: {issue}"
            
            return "Jira client initialized"
        except Exception as e:
            return f"Jira error: {str(e)}"


@dataclass
class NotionTool(BaseTool):
    """Notion integration tool."""
    
    name: str = "notion"
    description: str = "Interact with Notion. Input should be a JSON command."
    
    def execute(self, command_json: str) -> str:
        """Execute Notion command."""
        try:
            from open_agent.integrations.notion import NotionClient
            
            command = json.loads(command_json)
            client = NotionClient(command.get("token", ""))
            
            action = command.get("action", "list_pages")
            
            if action == "create_page":
                result = client.create_page(
                    parent_id=command.get("parent_id"),
                    title=command.get("title"),
                    content=command.get("content", "")
                )
                return f"Created page: {result}"
            elif action == "search":
                results = client.search(command.get("query", ""))
                return results
            
            return "Notion client ready"
        except Exception as e:
            return f"Notion error: {str(e)}"


@dataclass
class NewsTool(BaseTool):
    """Get latest news using News API."""
    
    name: str = "news"
    description: str = "Get latest news articles. Input should be a topic or 'latest'."
    
    def execute(self, topic: str = "latest") -> str:
        """Get news."""
        try:
            import httpx
            
            api_key = ""
            url = f"https://newsapi.org/v2/everything?q={topic}&apiKey={api_key}&pageSize=5"
            
            if not api_key:
                return "News API key not configured. Set NEWS_API_KEY in environment."
            
            response = httpx.get(url, timeout=10)
            data = response.json()
            
            articles = data.get("articles", [])
            if not articles:
                return "No news found."
            
            result = "Latest News:\n\n"
            for i, article in enumerate(articles[:5], 1):
                result += f"{i}. {article['title']}\n"
                result += f"   Source: {article['source']['name']}\n"
                result += f"   URL: {article['url']}\n\n"
            
            return result
        except Exception as e:
            return f"News error: {str(e)}"


def get_default_tools() -> List[BaseTool]:
    """Get a list of all default tools."""
    return [
        CalculatorTool(),
        WikipediaTool(),
        WebFetchTool(),
        PythonREPLTool(),
        NewsTool(),
    ]


def get_tools_by_category(category: str) -> List[BaseTool]:
    """Get tools by category."""
    categories = {
        "math": [CalculatorTool()],
        "research": [SearchTool(), WikipediaTool(), NewsTool()],
        "web": [WebFetchTool(), SearchTool()],
        "code": [PythonREPLTool()],
        "files": [FileReadTool(), FileWriteTool()],
        "productivity": [NotionTool(), JiraTool(), GitHubTool()],
    }
    return categories.get(category, [])



"""Gateway module - Message routing and session management.

This module provides the gateway system similar to OpenClaw:
- Protocol handling for agent communication
- Session management
- Authentication and authorization
- Message routing
"""

from open_agent.gateway.protocol.router import MessageRouter
from open_agent.gateway.sessions.manager import SessionManager
from open_agent.gateway.auth.auth import AuthHandler

__all__ = [
    "MessageRouter",
    "SessionManager",
    "AuthHandler",
]



"""OpenShell backend integration for deep agents.

This module provides integration with NVIDIA OpenShell sandbox runtime,
enabling secure, isolated execution environments for AI agents.
"""

from typing import Optional, Dict, Any, List
from enum import Enum
import structlog

from open_agent.config.settings import get_config, OpenShellConfig

logger = structlog.get_logger(__name__)


class SandboxType(str, Enum):
    """Types of sandbox environments."""
    DOCKER = "docker"
    K3S = "k3s"
    AUTO = "auto"


class SandboxSession:
    """Represents an OpenShell sandbox session."""

    def __init__(
        self,
        session_id: str,
        backend: "OpenShellBackend",
        metadata: Optional[Dict[str, Any]] = None,
    ):
        self.session_id = session_id
        self.backend = backend
        self.metadata = metadata or {}

    def exec(self, command: str) -> Dict[str, Any]:
        """Execute a command in the sandbox."""
        return self.backend._execute_command(self.session_id, command)

    def write_file(self, path: str, content: str) -> Dict[str, Any]:
        """Write a file to the sandbox."""
        return self.backend._write_file(self.session_id, path, content)

    def read_file(self, path: str) -> str:
        """Read a file from the sandbox."""
        return self.backend._read_file(self.session_id, path)

    def terminate(self) -> None:
        """Terminate the sandbox session."""
        self.backend._terminate_session(self.session_id)


class OpenShellBackend:
    """OpenShell sandbox backend for deep agents.

    This backend provides:
    - Isolated execution environments (Docker/K3s containers)
    - Policy-enforced filesystem, process, and network access
    - Privacy router for data protection
    - Credential injection for agents

    Uses NVIDIA OpenShell for secure agent execution with YAML-based policies.
    """

    def __init__(
        self,
        sandbox_type: SandboxType = SandboxType.AUTO,
        policy_file: Optional[str] = None,
        credential_provider: str = "nvidia",
        root_dir: Optional[str] = None,
    ):
        self.config = get_config().openshell
        self.sandbox_type = sandbox_type
        self.policy_file = policy_file or self.config.policy_file
        self.credential_provider = credential_provider
        self.root_dir = root_dir
        self._sessions: Dict[str, SandboxSession] = {}
        self._initialized = False

    def initialize(self) -> None:
        """Initialize the OpenShell backend."""
        if self._initialized:
            return

        logger.info("initializing_openshell_backend", sandbox_type=self.sandbox_type)

        # Check if OpenShell CLI is available
        try:
            import subprocess
            result = subprocess.run(
                ["openshell", "--version"],
                capture_output=True,
                text=True,
            )
            logger.info("openshell_version", version=result.stdout.strip())
        except FileNotFoundError:
            logger.warning("openshell_not_found", message="OpenShell CLI not installed")

        self._initialized = True

    def create_sandbox(self, agent_type: str = "nemotron") -> SandboxSession:
        """Create a new sandbox session.

        Args:
            agent_type: Type of agent to run (e.g., 'nemotron', 'claude', 'opencode')

        Returns:
            SandboxSession instance
        """
        if not self._initialized:
            self.initialize()

        import uuid
        session_id = f"sandbox_{uuid.uuid4().hex[:12]}"

        logger.info("creating_sandbox", session_id=session_id, agent_type=agent_type)

        # Try to use openshell CLI
        try:
            import subprocess
            cmd = [
                "openshell", "sandbox", "create",
                "--from", agent_type,
            ]
            if self.config.remote_host:
                cmd.extend(["--remote", self.config.remote_host])

            result = subprocess.run(
                cmd,
                capture_output=True,
                text=True,
                timeout=60,
            )

            if result.returncode == 0:
                logger.info("sandbox_created", session_id=session_id, output=result.stdout)
            else:
                logger.warning("sandbox_creation_failed", error=result.stderr)

        except Exception as e:
            logger.error("sandbox_creation_error", error=str(e))

        session = SandboxSession(session_id=session_id, backend=self)
        self._sessions[session_id] = session
        return session

    def _execute_command(self, session_id: str, command: str) -> Dict[str, Any]:
        """Execute a command in a sandbox session."""
        logger.debug("executing_command", session_id=session_id, command=command)

        try:
            import subprocess
            result = subprocess.run(
                ["openshell", "exec", session_id, command],
                capture_output=True,
                text=True,
                timeout=30,
            )

            return {
                "success": result.returncode == 0,
                "stdout": result.stdout,
                "stderr": result.stderr,
                "returncode": result.returncode,
            }

        except Exception as e:
            return {
                "success": False,
                "error": str(e),
                "stdout": "",
                "stderr": "",
                "returncode": -1,
            }

    def _write_file(self, session_id: str, path: str, content: str) -> Dict[str, Any]:
        """Write a file to the sandbox."""
        import base64
        content_b64 = base64.b64encode(content.encode()).decode()

        try:
            import subprocess
            result = subprocess.run(
                ["openshell", "file", "write", session_id, path, content_b64],
                capture_output=True,
                text=True,
                timeout=10,
            )

            return {
                "success": result.returncode == 0,
                "path": path,
            }

        except Exception as e:
            return {
                "success": False,
                "error": str(e),
            }

    def _read_file(self, session_id: str, path: str) -> str:
        """Read a file from the sandbox."""
        try:
            import subprocess
            result = subprocess.run(
                ["openshell", "file", "read", session_id, path],
                capture_output=True,
                text=True,
                timeout=10,
            )

            if result.returncode == 0:
                import base64
                return base64.b64decode(result.stdout).decode()

        except Exception as e:
            logger.error("read_file_error", session_id=session_id, path=path, error=str(e))

        return ""

    def _terminate_session(self, session_id: str) -> None:
        """Terminate a sandbox session."""
        if session_id in self._sessions:
            try:
                import subprocess
                subprocess.run(
                    ["openshell", "sandbox", "stop", session_id],
                    capture_output=True,
                    timeout=10,
                )
            except Exception as e:
                logger.error("terminate_session_error", session_id=session_id, error=str(e))

            del self._sessions[session_id]
            logger.info("session_terminated", session_id=session_id)

    def list_sessions(self) -> List[str]:
        """List active sandbox sessions."""
        return list(self._sessions.keys())

    def get_session(self, session_id: str) -> Optional[SandboxSession]:
        """Get a sandbox session by ID."""
        return self._sessions.get(session_id)


def create_openshell_backend(
    sandbox_type: str = "auto",
    policy_file: Optional[str] = None,
) -> OpenShellBackend:
    """Create an OpenShell backend.

    Example:
        >>> from open_agent.backends.openshell import create_openshell_backend
        >>>
        >>> backend = create_openshell_backend(
        ...     sandbox_type="docker",
        ...     policy_file="./policy.yaml"
        ... )
        >>> session = backend.create_sandbox("nemotron")
    """
    return OpenShellBackend(
        sandbox_type=SandboxType(sandbox_type),
        policy_file=policy_file,
    )


# Global backend instance
_global_backend: Optional[OpenShellBackend] = None


def get_default_backend() -> OpenShellBackend:
    """Get or create the default OpenShell backend."""
    global _global_backend
    if _global_backend is None:
        _global_backend = create_openshell_backend()
    return _global_backend



"""
Example 1: Basic Agent - Single AI assistant using NVIDIA Nemotron
"""
import asyncio
from open_agent import create_agent


async def main():
    print("=" * 60)
    print("Example 1: Basic Agent with NVIDIA Nemotron")
    print("=" * 60)
    
    agent = await create_agent(
        name="assistant",
        role="AI Assistant",
        goal="Help users with their questions accurately and efficiently",
        backstory="""You are a highly capable AI assistant. You have access to tools
        for calculations, web search, and Python code execution. Use them when needed
        to provide accurate answers.""",
        provider="nvidia",
    )
    
    print("\nAgent ready! Ask me anything...")
    
    while True:
        question = input("\nYou: ")
        if question.lower() in ['exit', 'quit', 'q']:
            break
        
        result = await agent.execute(question)
        print(f"\nAssistant: {result}")


if __name__ == "__main__":
    asyncio.run(main())



"""GitHub integration for OpenAgent.

Real API integration with GitHub for issues, PRs, repos, and more.
"""

from typing import Optional, Dict, List, Any
from dataclasses import dataclass
import structlog

logger = structlog.get_logger(__name__)


@dataclass
class GitHubConfig:
    """GitHub API configuration."""
    token: str = ""
    username: str = ""
    owner: str = ""


class GitHubClient:
    """GitHub API client."""

    def __init__(self, config: GitHubConfig):
        self.config = config
        self.base_url = "https://api.github.com"
        self._session = None

    async def initialize(self) -> bool:
        """Initialize GitHub client."""
        try:
            import httpx
            
            self._session = httpx.AsyncClient(
                headers={
                    "Authorization": f"Bearer {self.config.token}",
                    "Accept": "application/vnd.github.v3+json",
                    "X-GitHub-Api-Version": "2022-11-28",
                },
                timeout=30.0,
            )
            logger.info("github_initialized")
            return True
        except Exception as e:
            logger.error("github_init_failed", error=str(e))
            return False

    async def close(self):
        """Close the client."""
        if self._session:
            await self._session.aclose()

    async def _get(self, path: str) -> Optional[Dict]:
        """Make GET request."""
        try:
            response = await self._session.get(f"{self.base_url}{path}")
            response.raise_for_status()
            return response.json()
        except Exception as e:
            logger.error("github_request_failed", path=path, error=str(e))
            return None

    async def _post(self, path: str, data: Dict) -> Optional[Dict]:
        """Make POST request."""
        try:
            response = await self._session.post(f"{self.base_url}{path}", json=data)
            response.raise_for_status()
            return response.json()
        except Exception as e:
            logger.error("github_post_failed", path=path, error=str(e))
            return None

    async def get_user(self) -> Optional[Dict]:
        """Get current user info."""
        return await self._get("/user")

    async def get_repo(self, owner: str, repo: str) -> Optional[Dict]:
        """Get repository info."""
        return await self._get(f"/repos/{owner}/{repo}")

    async def list_repos(self, per_page: int = 30) -> List[Dict]:
        """List user repositories."""
        data = await self._get(f"/user/repos?per_page={per_page}")
        return data if isinstance(data, list) else []

    async def list_issues(
        self,
        owner: str,
        repo: str,
        state: str = "open",
        per_page: int = 30,
    ) -> List[Dict]:
        """List issues in a repository."""
        data = await self._get(f"/repos/{owner}/{repo}/issues?state={state}&per_page={per_page}")
        return [i for i in (data or []) if "pull_request" not in i]  # Filter out PRs

    async def create_issue(
        self,
        owner: str,
        repo: str,
        title: str,
        body: str = "",
        labels: Optional[List[str]] = None,
    ) -> Optional[Dict]:
        """Create an issue."""
        data = {
            "title": title,
            "body": body,
        }
        if labels:
            data["labels"] = labels
        return await self._post(f"/repos/{owner}/{repo}/issues", data)

    async def get_issue(self, owner: str, repo: str, issue_number: int) -> Optional[Dict]:
        """Get issue details."""
        return await self._get(f"/repos/{owner}/{repo}/issues/{issue_number}")

    async def update_issue(
        self,
        owner: str,
        repo: str,
        issue_number: int,
        title: Optional[str] = None,
        body: Optional[str] = None,
        state: Optional[str] = None,
        labels: Optional[List[str]] = None,
    ) -> Optional[Dict]:
        """Update an issue."""
        data = {}
        if title is not None:
            data["title"] = title
        if body is not None:
            data["body"] = body
        if state is not None:
            data["state"] = state
        if labels is not None:
            data["labels"] = labels

        try:
            response = await self._session.patch(
                f"{self.base_url}/repos/{owner}/{repo}/issues/{issue_number}",
                json=data,
            )
            response.raise_for_status()
            return response.json()
        except Exception as e:
            logger.error("github_update_issue_failed", error=str(e))
            return None

    async def add_issue_comment(
        self,
        owner: str,
        repo: str,
        issue_number: int,
        body: str,
    ) -> Optional[Dict]:
        """Add a comment to an issue."""
        return await self._post(
            f"/repos/{owner}/{repo}/issues/{issue_number}/comments",
            {"body": body},
        )

    async def list_pulls(
        self,
        owner: str,
        repo: str,
        state: str = "open",
    ) -> List[Dict]:
        """List pull requests."""
        return await self._get(f"/repos/{owner}/{repo}/pulls?state={state}") or []

    async def create_pull_request(
        self,
        owner: str,
        repo: str,
        title: str,
        head: str,
        base: str,
        body: str = "",
    ) -> Optional[Dict]:
        """Create a pull request."""
        data = {
            "title": title,
            "head": head,
            "base": base,
            "body": body,
        }
        return await self._post(f"/repos/{owner}/{repo}/pulls", data)

    async def get_pull_request(self, owner: str, repo: str, pr_number: int) -> Optional[Dict]:
        """Get PR details."""
        return await self._get(f"/repos/{owner}/{repo}/pulls/{pr_number}")

    async def merge_pull_request(
        self,
        owner: str,
        repo: str,
        pr_number: int,
        commit_title: Optional[str] = None,
        commit_message: Optional[str] = None,
    ) -> Optional[Dict]:
        """Merge a pull request."""
        data = {}
        if commit_title:
            data["commit_title"] = commit_title
        if commit_message:
            data["commit_message"] = commit_message

        try:
            response = await self._session.put(
                f"{self.base_url}/repos/{owner}/{repo}/merges",
                json=data,
            )
            response.raise_for_status()
            return response.json()
        except Exception as e:
            logger.error("github_merge_failed", error=str(e))
            return None

    async def list_commits(
        self,
        owner: str,
        repo: str,
        per_page: int = 30,
    ) -> List[Dict]:
        """List commits in a repository."""
        data = await self._get(f"/repos/{owner}/{repo}/commits?per_page={per_page}")
        return data if isinstance(data, list) else []

    async def get_commit(self, owner: str, repo: str, sha: str) -> Optional[Dict]:
        """Get commit details."""
        return await self._get(f"/repos/{owner}/{repo}/commits/{sha}")

    async def list_branches(self, owner: str, repo: str) -> List[Dict]:
        """List branches in a repository."""
        data = await self._get(f"/repos/{owner}/{repo}/branches")
        return data if isinstance(data, list) else []

    async def create_branch(
        self,
        owner: str,
        repo: str,
        branch: str,
        from_branch: str = "main",
    ) -> Optional[Dict]:
        """Create a new branch."""
        try:
            # Get the SHA of the source branch
            ref_data = await self._get(f"/repos/{owner}/{repo}/git/ref/heads/{from_branch}")
            if not ref_data:
                return None

            sha = ref_data.get("object", {}).get("sha")
            if not sha:
                return None

            # Create the new branch
            response = await self._session.post(
                f"{self.base_url}/repos/{owner}/{repo}/git/refs",
                json={
                    "ref": f"refs/heads/{branch}",
                    "sha": sha,
                },
            )
            response.raise_for_status()
            return response.json()
        except Exception as e:
            logger.error("github_create_branch_failed", error=str(e))
            return None

    async def create_file(
        self,
        owner: str,
        repo: str,
        path: str,
        content: str,
        message: str,
        branch: str = "main",
    ) -> Optional[Dict]:
        """Create or update a file."""
        import base64

        try:
            data = {
                "message": message,
                "content": base64.b64encode(content.encode()).decode(),
                "branch": branch,
            }
            return await self._put(f"/repos/{owner}/{repo}/contents/{path}", data)
        except Exception as e:
            logger.error("github_create_file_failed", error=str(e))
            return None

    async def get_file_content(
        self,
        owner: str,
        repo: str,
        path: str,
        ref: Optional[str] = None,
    ) -> Optional[str]:
        """Get file content."""
        import base64

        try:
            path_param = f"/repos/{owner}/{repo}/contents/{path}"
            if ref:
                path_param += f"?ref={ref}"

            data = await self._get(path_param)
            if data and "content" in data:
                content = data["content"].replace("\n", "")
                return base64.b64decode(content).decode()
        except Exception as e:
            logger.error("github_get_file_failed", error=str(e))
        return None

    async def _put(self, path: str, data: Dict) -> Optional[Dict]:
        """Make PUT request."""
        try:
            response = await self._session.put(f"{self.base_url}{path}", json=data)
            response.raise_for_status()
            return response.json()
        except Exception as e:
            logger.error("github_put_failed", path=path, error=str(e))
            return None

    async def search_repos(self, query: str, per_page: int = 30) -> List[Dict]:
        """Search repositories."""
        import httpx
        encoded_query = httpx.QueryParams({"q": query, "per_page": str(per_page)})
        
        try:
            response = await self._session.get(
                f"{self.base_url}/search/repositories",
                params=encoded_query,
            )
            response.raise_for_status()
            data = response.json()
            return data.get("items", [])
        except Exception as e:
            logger.error("github_search_failed", error=str(e))
            return []

    async def search_issues(self, query: str, per_page: int = 30) -> List[Dict]:
        """Search issues and PRs."""
        import httpx
        encoded_query = httpx.QueryParams({"q": query, "per_page": str(per_page)})
        
        try:
            response = await self._session.get(
                f"{self.base_url}/search/issues",
                params=encoded_query,
            )
            response.raise_for_status()
            data = response.json()
            return data.get("items", [])
        except Exception as e:
            logger.error("github_search_issues_failed", error=str(e))
            return []


def create_github_client(token: str, username: str = "") -> GitHubClient:
    """Create a GitHub client."""
    config = GitHubConfig(token=token, username=username)
    return GitHubClient(config)



"""Skills module - Loadable agent capabilities from markdown files.

This module provides a skills system similar to OpenClaw:
- Skills defined in markdown files
- Automatic skill loading from workspace
- Skill invocation and context
- Skill discovery and management
"""

from open_agent.agents.skills.registry import Skill, SkillRegistry, SkillInvocation

__all__ = [
    "Skill",
    "SkillRegistry", 
    "SkillInvocation",
]
